import os
import sys
from typing import Any, TypeGuard

from tomlkit import document, table
from tomlkit.items import AoT, Table
from tomlkit.toml_document import TOMLDocument

from ..i18n import _
from ..log import RuyiLogger
from . import checksum
from .pkg_manifest import DistfileDeclType, RestrictKind


def do_admin_checksum(
    logger: RuyiLogger,
    files: list[os.PathLike[Any]],
    format: str,
    restrict: list[str],
) -> int:
    if not validate_restrict_kinds(restrict):
        logger.F(
            _("invalid restrict kinds given: {restrict}").format(restrict=restrict)
        )
        return 1

    entries = [gen_distfile_entry(logger, f, restrict) for f in files]
    if format == "toml":
        doc = emit_toml_distfiles_section(entries)
        logger.D(f"{doc}")
        sys.stdout.write(doc.as_string())
        return 0

    raise RuntimeError("unrecognized output format; should never happen")


def validate_restrict_kinds(input: list[str]) -> TypeGuard[list[RestrictKind]]:
    for x in input:
        match x:
            case "fetch" | "mirror":
                pass
            case _:
                return False
    return True


def gen_distfile_entry(
    logger: RuyiLogger,
    path: os.PathLike[Any],
    restrict: list[RestrictKind],
) -> DistfileDeclType:
    logger.D(f"generating distfile entry for {path}")
    with open(path, "rb") as fp:
        filesize = os.stat(fp.fileno()).st_size
        c = checksum.Checksummer(fp, {})
        checksums = c.compute(kinds=checksum.SUPPORTED_CHECKSUM_KINDS)

    obj: DistfileDeclType = {
        "name": os.path.basename(path),
        "size": filesize,
        "checksums": checksums,
    }

    if restrict:
        obj["restrict"] = restrict

    return obj


def emit_toml_distfiles_section(x: list[DistfileDeclType]) -> TOMLDocument:
    doc = document()

    arr: list[Table] = []
    for dd in x:
        t = table()
        t.add("name", dd["name"])
        t.add("size", dd["size"])
        if r := dd.get("restrict"):
            t.add("restrict", r)
        t.add("checksums", emit_toml_checksums(dd["checksums"]))
        arr.append(t)

    doc.add("distfiles", AoT(arr))
    return doc


def emit_toml_checksums(x: dict[str, str]) -> Table:
    t = table()
    for k in sorted(x.keys()):
        t.add(k, x[k])
    return t



import abc
import json
import os
import pathlib
import tomllib
from typing import Any, Mapping, Sequence

from ..i18n import _
from ..log import RuyiLogger
from ..utils.porcelain import PorcelainEntity, PorcelainEntityType


class PorcelainEntityListOutputV1(PorcelainEntity):
    entity_type: str
    entity_id: str
    display_name: str | None
    data: Mapping[str, Any]
    related_refs: list[str]
    reverse_refs: list[str]


class EntityError(Exception):
    """Base exception for entity-related errors."""

    pass


class EntityValidationError(EntityError):
    """Exception raised when an entity fails validation."""

    def __init__(self, entity_type: str, entity_id: str, cause: Exception) -> None:
        self.entity_type = entity_type
        self.entity_id = entity_id
        self.cause = cause
        message = (
            f"Entity validation failed for entity '{entity_type}:{entity_id}': {cause}"
        )
        super().__init__(message)


class BaseEntity:
    """Base class for all entity types."""

    def __init__(
        self,
        entity_type: str,
        entity_id: str,
        data: Mapping[str, Any],
    ) -> None:
        self._entity_type = entity_type
        self._id = entity_id
        self._data = data

        self._reverse_refs: set[str] = set()

    @property
    def entity_type(self) -> str:
        """Type of the entity."""
        return self._entity_type

    @property
    def id(self) -> str:
        """ID of the entity."""
        return self._id

    @property
    def display_name(self) -> str | None:
        """Human-readable name of the entity."""
        result = self._data[self.entity_type].get("display_name", None)
        if result is None or isinstance(result, str):
            return result
        # return None if type is unexpected
        return None

    @property
    def unique_among_type_during_traversal(self) -> bool:
        """Whether the entity should be unique among all entities of the same type
        during traversal.

        For example, if the entity is ``arch:foo64`` and there is also ``arch:foo32``,
        with this property set to ``True`` on each, there will be only one
        ``arch:foo*`` entity in any traversal path involving them, so that a
        hypothetical traversal starting from a "foo64" device will not return
        entities only related to the "foo32" architecture.
        """

        if r := self._data.get("unique_among_type_during_traversal", None):
            if isinstance(r, bool):
                return r
        # return False if type is unexpected
        return False

    @property
    def data(self) -> Any:
        """Raw data of the entity."""
        return self._data[self.entity_type]

    @property
    def related_refs(self) -> list[str]:
        """Get the list of related entity references."""
        if r := self._data.get("related"):
            if isinstance(r, list):
                return r
        # return empty list if that is the case, or if the type is unexpected
        return []

    @property
    def reverse_refs(self) -> list[str]:
        """Get the list of reverse-related entity references."""
        return list(self._reverse_refs)

    def _add_reverse_ref(self, ref: str) -> None:
        self._reverse_refs.add(ref)

    def to_porcelain(self) -> PorcelainEntityListOutputV1:
        """Convert this entity to porcelain output format."""

        return {
            "ty": PorcelainEntityType.EntityListOutputV1,
            "entity_type": self.entity_type,
            "entity_id": self.id,
            "display_name": self.display_name,
            "data": self._data,
            "related_refs": self.related_refs,
            "reverse_refs": self.reverse_refs,
        }

    def __str__(self) -> str:
        return f"{self.entity_type}:{self.id}"

    def __hash__(self) -> int:
        return hash((self.entity_type, self.id))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseEntity):
            return NotImplemented
        return self.entity_type == other.entity_type and self.id == other.id


class BaseEntityProvider(abc.ABC):
    """Abstract base class for entity data providers.

    Entity providers are responsible for discovering and loading entity schemas and data.
    """

    @abc.abstractmethod
    def discover_schemas(self) -> dict[str, object]:
        """Discover available entity schemas.

        Returns:
            A dictionary mapping entity types to their schema objects
        """
        raise NotImplementedError

    @abc.abstractmethod
    def load_entities(
        self,
        entity_types: Sequence[str],
    ) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
        """Load entities of the given types.

        Args:
            entity_types: Sequence of entity types to load

        Returns:
            A nested dictionary mapping entity types to entity IDs to raw entity data
        """
        raise NotImplementedError


class FSEntityProvider(BaseEntityProvider):
    """Entity provider that loads entity data from the filesystem.

    This provider reads schemas from the ``_schemas`` directory and entity data from
    subdirectories organized by entity type.
    """

    def __init__(self, logger: RuyiLogger, entities_root: os.PathLike[Any]) -> None:
        """Initialize the filesystem-based entity provider.

        Args:
            logger: Logger instance to use.
            entities_root: Path to the root directory containing entity data.
                           The ``_schemas`` directory should be a subdirectory of this path.
        """

        self._logger = logger
        self._entities_root = pathlib.Path(entities_root)
        self._schemas_root = self._entities_root / "_schemas"

    def discover_schemas(self) -> dict[str, object]:
        """Discover entity schemas from the filesystem.

        Returns:
            A dictionary mapping entity types to their schema objects
        """
        schemas: dict[str, object] = {}

        if not os.path.isdir(self._schemas_root):
            self._logger.D(f"entity schemas directory not found: {self._schemas_root}")
            return schemas

        try:
            schema_files = list(self._schemas_root.glob("*.jsonschema"))
        except IOError as e:
            self._logger.W(
                _("failed to access entity schemas directory {dir}: {reason}").format(
                    dir=self._schemas_root,
                    reason=e,
                )
            )
            return schemas

        for p in schema_files:
            # Extract entity type from schema filename (remove .jsonschema extension)
            entity_type = p.name[:-11]  # 11 is the length of ".jsonschema"

            try:
                with open(p, "r", encoding="utf-8") as f:
                    schema = json.load(f)
            except (IOError, json.JSONDecodeError) as e:
                self._logger.D(
                    f"failed to load schema for entity type '{entity_type}': {e}"
                )
                continue

            # Cache the schema
            schemas[entity_type] = schema

        self._logger.D(f"discovered entity types from schemas: {list(schemas.keys())}")
        return schemas

    def load_entities(
        self,
        entity_types: Sequence[str],
    ) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
        """Load entity data from the filesystem.

        Args:
            entity_types: Set of entity types to load

        Returns:
            A nested dictionary mapping entity types to entity IDs to raw entity data
        """
        entities: dict[str, dict[str, dict[str, Any]]] = {
            entity_type: {} for entity_type in entity_types
        }

        for entity_type in entity_types:
            type_dir = self._entities_root / entity_type

            if not type_dir.exists():
                self._logger.D(f"entity type directory does not exist: {type_dir}")
                continue

            for file_path in type_dir.glob("*.toml"):
                try:
                    with open(file_path, "rb") as f:
                        data = tomllib.load(f)
                except (IOError, tomllib.TOMLDecodeError) as e:
                    self._logger.W(
                        _("failed to load entity from {path}: {reason}").format(
                            path=file_path,
                            reason=e,
                        )
                    )
                    continue

                # Extract entity ID from filename (remove .toml extension)
                entity_id = file_path.name[:-5]

                # Create and store raw entity data
                entities[entity_type][entity_id] = data

        entity_counts = {t: len(e) for t, e in entities.items()}
        self._logger.D(f"count of loaded entities from filesystem: {entity_counts}")
        return entities



import argparse
from typing import TYPE_CHECKING

from ..i18n import _
from .cmd import RootCommand

if TYPE_CHECKING:
    from .completion import ArgumentParser
    from ..config import GlobalConfig


class VersionCommand(
    RootCommand,
    cmd="version",
    help=_("Print version information"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        return cli_version(cfg, args)


def cli_version(cfg: "GlobalConfig", args: argparse.Namespace) -> int:
    import ruyi
    from ..ruyipkg.host import get_native_host
    from ..version import COPYRIGHT_NOTICE, MPL_REDIST_NOTICE, RUYI_SEMVER

    print(
        _("Ruyi {version}\n\nRunning on {host}.").format(
            version=RUYI_SEMVER,
            host=get_native_host(),
        )
    )

    if cfg.is_installation_externally_managed:
        print(_("This Ruyi installation is externally managed."))

    print()

    cfg.logger.stdout(_(COPYRIGHT_NOTICE))

    # Output the MPL notice only when we actually bundle and depend on the
    # MPL component(s), which right now is only certifi. Keep the condition
    # synced with __main__.py.
    if hasattr(ruyi, "__compiled__") and ruyi.__compiled__.standalone:
        cfg.logger.stdout(_(MPL_REDIST_NOTICE))

    return 0



from ruyi.utils.global_mode import is_cli_completion_script_requested


def test_is_cli_completion_script_requested_matches_only_exact_option() -> None:
    assert is_cli_completion_script_requested(["ruyi", "--output-completion-script"])
    assert is_cli_completion_script_requested(
        ["ruyi", "--output-completion-script=bash"]
    )

    assert not is_cli_completion_script_requested(
        ["ruyi", "--output-completion-script-foo"]
    )
    assert not is_cli_completion_script_requested(
        ["ruyi", "--output-completion-script-foo=bash"]
    )



from typing import TypedDict, TypeGuard, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import NotRequired


class TelemetryEvent(TypedDict):
    fmt: int
    time_bucket: "NotRequired[str]"  # canonically "YYYYMMDDHHMM"
    kind: str
    params: dict[str, object]


def is_telemetry_event(x: object) -> TypeGuard[TelemetryEvent]:
    if not isinstance(x, dict):
        return False

    if not 3 <= len(x.keys()) <= 4:
        return False

    try:
        if not isinstance(x["fmt"], int):
            return False
        if not isinstance(x["kind"], str):
            return False
        if not isinstance(x["params"], dict):
            return False
    except KeyError:
        return False

    try:
        if not isinstance(x["time_bucket"], str):
            return False
        if len(x["time_bucket"]) != 12:
            return False
        if not x["time_bucket"].isdigit():
            return False
    except KeyError:
        pass

    return True



import os
import pathlib
import shutil
from typing import TYPE_CHECKING

from ..i18n import _

if TYPE_CHECKING:
    from ..log import RuyiLogger


def migrate_repo_dir(cache_root: str, logger: "RuyiLogger") -> None:
    """Migrate the legacy packages-index/ directory to repos/ruyisdk/.

    If ``<cache>/packages-index/`` exists (not as a symlink) and
    ``<cache>/repos/ruyisdk/`` does not, move the former to the latter
    and create a compatibility symlink at the old location.

    If both exist, or the old path is already a symlink, do nothing.
    """

    legacy_path = pathlib.Path(cache_root) / "packages-index"
    new_path = pathlib.Path(cache_root) / "repos" / "ruyisdk"

    # Nothing to migrate if the legacy directory doesn't exist or is
    # already a symlink (i.e. a previous migration already ran).
    if not legacy_path.exists() or legacy_path.is_symlink():
        return

    # Already migrated (both exist) — do nothing.
    if new_path.exists():
        return

    logger.I(
        _("migrating repo directory from [yellow]{old}[/] to [yellow]{new}[/]").format(
            old=legacy_path, new=new_path
        )
    )

    new_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.move(str(legacy_path), str(new_path))

    # Create a compatibility symlink so that any code still using the old
    # path continues to work during the transition.
    os.symlink(str(new_path), str(legacy_path))

    logger.I(_("repo directory migration complete"))



from ruyi.utils.l10n import lang_code_to_lang_region, LangAndRegion, match_lang_code


def test_lang_code_to_lang_region() -> None:
    assert lang_code_to_lang_region("en", False) == LangAndRegion("en", "en", None)
    assert lang_code_to_lang_region("en", True) == LangAndRegion("en", "en", "US")
    assert lang_code_to_lang_region("en_SG", False) == LangAndRegion(
        "en_SG", "en", "SG"
    )
    assert lang_code_to_lang_region("zh", False) == LangAndRegion("zh", "zh", None)
    assert lang_code_to_lang_region("zh", True) == LangAndRegion("zh", "zh", "CN")
    assert lang_code_to_lang_region("zh_HK", False) == LangAndRegion(
        "zh_HK", "zh", "HK"
    )
    assert lang_code_to_lang_region("cmn", False) == LangAndRegion("cmn", "cmn", None)
    assert lang_code_to_lang_region("cmn", True) == LangAndRegion("cmn", "cmn", None)


def test_match_lang_code() -> None:
    assert match_lang_code("zh", ["en"]) == "en"
    assert match_lang_code("en_US", ["en"]) == "en"
    assert match_lang_code("en", ["en_US"]) == "en_US"
    assert match_lang_code("en_US", ["en_US"]) == "en_US"

    assert match_lang_code("zh", ["en", "zh"]) == "zh"
    assert match_lang_code("zh", ["en", "zh_CN"]) == "zh_CN"
    assert match_lang_code("zh", ["en", "zh_HK"]) == "zh_HK"
    assert match_lang_code("zh_HK", ["en", "zh_CN"]) == "zh_CN"
    assert match_lang_code("zh_CN", ["en", "zh_HK"]) == "zh_HK"

    # match according to region
    assert match_lang_code("ga", ["en", "en_IE", "zh_CN"]) == "en_IE"

    # match according to language
    assert match_lang_code("pt", ["pt_BR", "en", "zh"]) == "pt_BR"

    # fallback in the order of en_US, en_*, zh_CN, zh_*
    assert (
        match_lang_code("pt", ["ga", "zh_HK", "zh", "zh_CN", "en_IE", "en", "en_US"])
        == "en_US"
    )
    assert match_lang_code("pt", ["ga", "zh_HK", "zh", "zh_CN", "en_IE", "en"]) == "en"
    assert match_lang_code("pt", ["ga", "zh_HK", "zh", "zh_CN", "en_IE"]) == "en_IE"
    assert match_lang_code("pt", ["ga", "zh_HK", "zh", "zh_CN"]) == "zh_CN"
    assert match_lang_code("pt", ["ga", "zh_HK", "zh"]) == "zh"
    assert match_lang_code("pt", ["ga", "zh_HK"]) == "zh_HK"

    # fallback to the lexicographically first one
    assert match_lang_code("ru", ["ga", "es_ES"]) == "es_ES"
    assert match_lang_code("ru", ["es_ES", "ga"]) == "es_ES"



import hashlib
from typing import BinaryIO, Final, Iterable

from ..i18n import _

SUPPORTED_CHECKSUM_KINDS: Final = {"sha256", "sha512"}


def get_hash_instance(kind: str) -> "hashlib._Hash":
    if kind not in SUPPORTED_CHECKSUM_KINDS:
        raise ValueError(_("checksum algorithm {kind} not supported").format(kind=kind))
    return hashlib.new(kind)


class Checksummer:
    def __init__(self, file: BinaryIO, checksums: dict[str, str]) -> None:
        self.file = file
        self.checksums = checksums

    def check(self) -> None:
        computed_csums = self.compute()
        for kind, expected_csum in self.checksums.items():
            if computed_csums[kind] != expected_csum:
                raise ValueError(
                    _("wrong {kind} checksum: want {want}, got {got}").format(
                        kind=kind,
                        want=expected_csum,
                        got=computed_csums[kind],
                    )
                )

    def compute(
        self,
        kinds: Iterable[str] | None = None,
        chunksize: int = 4096,
    ) -> dict[str, str]:
        if kinds is None:
            kinds = self.checksums.keys()

        checksummers = {kind: get_hash_instance(kind) for kind in kinds}
        while chunk := self.file.read(chunksize):
            for h in checksummers.values():
                h.update(chunk)

        return {kind: h.hexdigest() for kind, h in checksummers.items()}



import typing

if typing.TYPE_CHECKING:

    class NuitkaVersion(typing.NamedTuple):
        major: int
        minor: int
        micro: int
        releaselevel: str
        containing_dir: str
        standalone: bool
        onefile: bool
        macos_bundle_mode: bool
        no_asserts: bool
        no_docstrings: bool
        no_annotations: bool
        module: bool
        main: str
        original_argv0: str | None

    __compiled__: NuitkaVersion



import glob
from dataclasses import dataclass
from functools import cached_property
import itertools
import os
import os.path
import pathlib
import re
import tomllib
from typing import (
    Any,
    Final,
    Iterable,
    Mapping,
    Sequence,
    TypedDict,
    TypeGuard,
    TYPE_CHECKING,
    cast,
)
from urllib import parse

from pygit2 import clone_repository
from pygit2.repository import Repository

from ..i18n import _
from ..log import RuyiLogger
from ..pluginhost.ctx import PluginHostContext
from ..telemetry.scope import TelemetryScopeConfig
from ..utils.git import RemoteGitProgressIndicator, pull_ff_or_die
from ..utils.porcelain import PorcelainEntity, PorcelainEntityType
from ..utils.url import urljoin_for_sure
from .entity import EntityStore
from .entity_provider import BaseEntityProvider, FSEntityProvider
from .msg import RepoMessageStore
from .news_store import NewsItemStore
from .pkg_manifest import (
    BoundPackageManifest,
    DistfileDecl,
    InputPackageManifestType,
    is_prerelease,
)
from .profile import PluginProfileProvider, ProfileEntityProvider, ProfileProxy
from .protocols import ProvidesPackageManifests

if TYPE_CHECKING:
    from typing_extensions import NotRequired

    # for avoiding circular import
    from ..config import GlobalConfig


REPO_ID_PATTERN: Final = re.compile(r"^[a-z0-9][a-z0-9_-]*$")
DEFAULT_REPO_ID: Final = "ruyisdk"
DEFAULT_REPO_NAME: Final = "RuyiSDK official repository"
DEFAULT_REPO_PRIORITY: Final = 0


@dataclass(frozen=True)
class RepoEntry:
    """A configured pointer to a single metadata repository."""

    id: str
    name: str
    remote: str | None
    branch: str
    local_path: str | None
    priority: int
    active: bool
    is_system: bool = False

    def resolve_root(self, cache_root: str | os.PathLike[str]) -> str:
        """Return the local checkout path for this repo entry."""
        if self.local_path is not None:
            return self.local_path
        return os.path.join(cache_root, "repos", self.id)

    @classmethod
    def from_legacy_config(cls, gc: "GlobalConfig") -> "RepoEntry":
        """Construct the default RepoEntry from the existing GlobalConfig
        repo fields (backward-compatible single-repo path)."""
        from ..config import DEFAULT_REPO_URL, DEFAULT_REPO_BRANCH

        return cls(
            id=DEFAULT_REPO_ID,
            name=DEFAULT_REPO_NAME,
            remote=gc.override_repo_url or DEFAULT_REPO_URL,
            branch=gc.override_repo_branch or DEFAULT_REPO_BRANCH,
            local_path=gc.override_repo_dir,
            priority=DEFAULT_REPO_PRIORITY,
            active=True,
        )

    def make_metadata_repo(self, gc: "GlobalConfig") -> "MetadataRepo":
        """Construct a MetadataRepo from this entry's fields."""
        root = self.resolve_root(str(gc.cache_root))
        return MetadataRepo(
            gc,
            root=root,
            remote=self.remote or "",
            branch=self.branch,
            repo_id=self.id,
            repo_name=self.name,
        )

    def to_porcelain(self) -> "PorcelainRepoEntryV1":
        return {
            "ty": PorcelainEntityType.RepoEntryV1,
            "id": self.id,
            "name": self.name,
            "remote": self.remote,
            "branch": self.branch,
            "local_path": self.local_path,
            "priority": self.priority,
            "active": self.active,
            "is_system": self.is_system,
        }


class RepoConfigV0Type(TypedDict):
    dist: str
    doc_uri: "NotRequired[str]"


def validate_repo_config_v0(x: object) -> TypeGuard[RepoConfigV0Type]:
    if not isinstance(x, dict):
        return False
    if "ruyi-repo" in x:
        return False
    if "dist" not in x or not isinstance(x["dist"], str):
        return False
    if "doc_uri" in x and not isinstance(x["doc_uri"], str):
        return False
    return True


class RepoConfigV1Repo(TypedDict):
    doc_uri: "NotRequired[str]"
    id: "NotRequired[str]"
    name: "NotRequired[str]"


class RepoConfigV1Mirror(TypedDict):
    id: str
    urls: list[str]


class RepoConfigV1Telemetry(TypedDict):
    id: str
    scope: TelemetryScopeConfig
    url: str


RepoConfigV1Type = TypedDict(
    "RepoConfigV1Type",
    {
        "ruyi-repo": str,
        "repo": "NotRequired[RepoConfigV1Repo]",
        "mirrors": list[RepoConfigV1Mirror],
        "telemetry": "NotRequired[list[RepoConfigV1Telemetry]]",
    },
)


def validate_repo_config_v1(x: object) -> TypeGuard[RepoConfigV1Type]:
    if not isinstance(x, dict):
        return False
    x = cast(dict[str, object], x)
    if x.get("ruyi-repo", "") != "v1":
        return False
    return True


MIRROR_ID_RUYI_DIST: Final = "ruyi-dist"


class RepoConfig:
    def __init__(
        self,
        mirrors: list[RepoConfigV1Mirror],
        repo: RepoConfigV1Repo | None,
        telemetry_apis: list[RepoConfigV1Telemetry] | None,
    ) -> None:
        self.mirrors = {x["id"]: x["urls"] for x in mirrors}
        self.repo = repo

        self.telemetry_apis: dict[str, RepoConfigV1Telemetry]
        if telemetry_apis is not None:
            self.telemetry_apis = {x["id"]: x for x in telemetry_apis}
        else:
            self.telemetry_apis = {}

    @classmethod
    def from_object(cls, obj: object) -> "RepoConfig":
        if not isinstance(obj, dict):
            raise ValueError("repo config must be a dict")
        if "ruyi-repo" in obj:
            return cls.from_v1(cast(object, obj))
        return cls.from_v0(cast(object, obj))

    @classmethod
    def from_v0(cls, obj: object) -> "RepoConfig":
        if not validate_repo_config_v0(obj):
            # TODO: more detail in the error message
            raise RuntimeError("malformed v0 repo config")

        v1_mirrors: list[RepoConfigV1Mirror] = [
            {
                "id": MIRROR_ID_RUYI_DIST,
                "urls": [urljoin_for_sure(obj["dist"], "dist/")],
            },
        ]

        v1_repo: RepoConfigV1Repo | None = None
        if "doc_uri" in obj:
            v1_repo = {
                "doc_uri": obj["doc_uri"],
            }

        return cls(v1_mirrors, v1_repo, None)

    @classmethod
    def from_v1(cls, obj: object) -> "RepoConfig":
        if not validate_repo_config_v1(obj):
            # TODO: more detail in the error message
            raise RuntimeError("malformed v1 repo config")
        return cls(obj["mirrors"], obj.get("repo"), obj.get("telemetry"))

    @property
    def repo_id(self) -> str:
        if self.repo is not None and "id" in self.repo:
            return self.repo["id"]
        return "ruyisdk"

    @property
    def repo_name(self) -> str:
        if self.repo is not None and "name" in self.repo:
            return self.repo["name"]
        return "RuyiSDK official repository"

    def get_dist_urls_for_file(self, logger: RuyiLogger, url: str) -> list[str]:
        u = parse.urlparse(url)
        path = u.path.lstrip("/")
        match u.scheme:
            case "":
                return self.get_mirror_urls_for_file(MIRROR_ID_RUYI_DIST, path)
            case "mirror":
                return self.get_mirror_urls_for_file(u.netloc, path)
            case "http" | "https":
                # pass-through known protocols
                return [url]
            case _:
                # deny others
                logger.W(
                    _("unrecognized dist URL scheme: {scheme}").format(scheme=u.scheme)
                )
                return []

    def get_mirror_urls_for_file(self, mirror_id: str, path: str) -> list[str]:
        mirror_urls = self.mirrors.get(mirror_id, [])
        return [parse.urljoin(base, path) for base in mirror_urls]

    def get_telemetry_api_url(self, scope: TelemetryScopeConfig) -> str | None:
        for api_decl in self.telemetry_apis.values():
            if api_decl.get("scope", "") == scope:
                return api_decl.get("url", None)
        return None


class ArchProfileStore:
    def __init__(self, phctx: PluginHostContext[Any, Any], arch: str) -> None:
        self._arch = arch
        plugin_id = f"ruyi-profile-{arch}"
        self._provider = PluginProfileProvider(phctx, plugin_id)
        self._init_cache()

    def _init_cache(self) -> None:
        self._profiles_cache: dict[str, ProfileProxy] = {}
        for profile_id in self._provider.list_all_profile_ids():
            self._profiles_cache[profile_id] = ProfileProxy(
                self._provider, self._arch, profile_id
            )

    def __contains__(self, profile_id: str) -> bool:
        return profile_id in self._profiles_cache

    def __getitem__(self, profile_id: str) -> ProfileProxy:
        try:
            return self._profiles_cache[profile_id]
        except KeyError as e:
            raise KeyError(
                f"profile '{profile_id}' is not supported by this arch"
            ) from e

    def get(self, profile_id: str) -> ProfileProxy | None:
        return self._profiles_cache.get(profile_id)

    def iter_profiles(self) -> Iterable[ProfileProxy]:
        return self._profiles_cache.values()


class MetadataRepo(ProvidesPackageManifests):
    def __init__(
        self,
        gc: "GlobalConfig",
        *,
        root: str,
        remote: str,
        branch: str,
        repo_id: str = DEFAULT_REPO_ID,
        repo_name: str = DEFAULT_REPO_NAME,
    ) -> None:
        self._gc = gc
        self.root = root
        self.remote = remote
        self.branch = branch
        self._repo_id = repo_id
        self._repo_name = repo_name
        self.repo: Repository | None = None

        self._cfg: RepoConfig | None = None
        self._cfg_initialized = False
        self._pkgs: dict[str, dict[str, BoundPackageManifest]] = {}
        self._categories: dict[str, dict[str, dict[str, BoundPackageManifest]]] = {}
        self._slug_cache: dict[str, BoundPackageManifest] = {}
        self._supported_arches: set[str] | None = None
        self._arch_profile_stores: dict[str, ArchProfileStore] = {}
        self._news_cache: NewsItemStore | None = None
        self._plugin_host_ctx = PluginHostContext.new(
            gc.logger,
            self.plugin_root,
            locale=gc.lang_code,
            message_store_factory=lambda: self.messages,
        )
        self._plugin_fn_evaluator = self._plugin_host_ctx.make_evaluator()

    @classmethod
    def from_global_config(cls, gc: "GlobalConfig") -> "MetadataRepo":
        """Factory that preserves the current single-repo construction path.

        All existing call sites that used ``MetadataRepo(gc)`` should use
        this instead.

        .. deprecated::
            Use ``RepoEntry.make_metadata_repo()`` or the CompositeRepo
            via ``GlobalConfig.repo`` instead.
        """
        import warnings

        warnings.warn(
            "MetadataRepo.from_global_config() is deprecated; "
            "use RepoEntry.make_metadata_repo() or GlobalConfig.repo instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return cls(
            gc,
            root=gc.get_repo_dir(),
            remote=gc.get_repo_url(),
            branch=gc.get_repo_branch(),
        )

    @property
    def repo_id(self) -> str:
        return self._repo_id

    @property
    def repo_name(self) -> str:
        return self._repo_name

    @property
    def logger(self) -> RuyiLogger:
        return self._gc.logger

    @property
    def plugin_root(self) -> pathlib.Path:
        return pathlib.Path(self.root) / "plugins"

    def iter_plugin_ids(self) -> Iterable[str]:
        try:
            for p in self.plugin_root.iterdir():
                if p.is_dir():
                    yield p.name
        except (FileNotFoundError, NotADirectoryError):
            pass

    def get_from_plugin(self, plugin_id: str, key: str) -> object | None:
        return self._plugin_host_ctx.get_from_plugin(plugin_id, key)

    def eval_plugin_fn(
        self,
        function: object,
        *args: object,
        **kwargs: object,
    ) -> object:
        """Evaluates a function from a plugin.

        NOTE: There is security implication for the unsandboxed plugin backend,
        which provides **NO GUARDS** against arbitrary inputs for the ``function``
        argument because there is **no sandbox**."""

        return self._plugin_fn_evaluator.eval_function(function, *args, **kwargs)

    def ensure_git_repo(self) -> Repository:
        if self.repo is not None:
            return self.repo

        if os.path.exists(self.root):
            self.repo = Repository(self.root)
            return self.repo

        self.logger.I(
            _("the package repository does not exist at [yellow]{root}[/]").format(
                root=self.root
            )
        )
        self.logger.I(
            _("cloning from [cyan link={remote}]{remote}[/]").format(remote=self.remote)
        )

        with RemoteGitProgressIndicator() as pr:
            repo = clone_repository(
                self.remote,
                self.root,
                checkout_branch=self.branch,
                callbacks=pr,
            )
            # pygit2's type info is incomplete as of 1.16.0, and pyright
            # will not look at the typeshed stub for the appropriate signature
            # because pygit2 has the py.typed marker. Workaround the error for
            # now by explicitly casting to the right runtime type.
            self.repo = cast(Repository, repo)  # type: ignore[redundant-cast]

        # reinit config after cloning
        self._cfg_initialized = False
        self._read_config(False)

        return self.repo

    def sync(self) -> None:
        if not self.remote:
            # Local-only repo: no git remote to sync from; just reload metadata.
            self._gc.logger.D(
                _("skipping sync for local-only repo '{id}'").format(
                    id=self._repo_id,
                )
            )
            self._cfg_initialized = False
            self._read_config(False)
            return

        self._gc.logger.I(_("updating the package repository"))

        repo = self.ensure_git_repo()

        # only manage the repo settings on the user's behalf if the user
        # has not overridden the repo directory themselves
        allow_auto_management = not self._gc.have_overridden_repo_dir

        pull_ff_or_die(
            self.logger,
            repo,
            "origin",
            self.remote,
            self.branch,
            allow_auto_management=allow_auto_management,
        )

        self._gc.logger.I(_("package repository is updated"))

    @property
    def global_config(self) -> "GlobalConfig":
        return self._gc

    @property
    def config(self) -> RepoConfig:
        x = self._read_config(True)
        assert x is not None
        return x

    @property
    def maybe_config(self) -> RepoConfig | None:
        """Like ``config``, but does not pull down the repo in case the repo is
        not locally present at invocation time."""
        return self._read_config(False)

    def _read_config(self, ensure_if_not_existing: bool) -> RepoConfig | None:
        if self._cfg_initialized:
            return self._cfg

        if ensure_if_not_existing:
            self.ensure_git_repo()

        # we can read the config file directly because we're operating from a
        # working tree (as opposed to a bare repo)
        #
        # this is a fake loop (that "loops" only once)
        # here it's only for being able to use break's
        while True:
            try:
                with open(os.path.join(self.root, "config.toml"), "rb") as fp:
                    obj = tomllib.load(fp)
                    break
            except FileNotFoundError:
                pass

            self._cfg_initialized = True
            return None

        self._cfg_initialized = True
        self._cfg = RepoConfig.from_object(obj)
        return self._cfg

    @cached_property
    def messages(self) -> RepoMessageStore:
        self.ensure_git_repo()

        obj: dict[str, object] = {}
        try:
            with open(os.path.join(self.root, "messages.toml"), "rb") as fp:
                obj = tomllib.load(fp)
        except FileNotFoundError:
            pass

        return RepoMessageStore.from_object(obj)

    def iter_pkg_manifests(
        self,
        ensure_repo: bool = True,
    ) -> Iterable[BoundPackageManifest]:
        if ensure_repo:
            self.ensure_git_repo()

        # Try the name "packages" first, because it's possible that some
        # non-manifest things will arrive in the package directories in the
        # future. Keep probing the old "manifests" name until 0.51.0.
        #
        # TODO: remove the "manifests" alias after 0.50.0 branch is cut.
        for pkg_dir_name in ("packages", "manifests"):
            manifests_dir = os.path.join(self.root, pkg_dir_name)
            try:
                manifests_dir_iter = os.scandir(manifests_dir)
            except FileNotFoundError:
                continue

            for f in manifests_dir_iter:
                if not f.is_dir():
                    continue
                yield from self._iter_pkg_manifests_from_category(f.path)
            # Only process the first matched directory
            return

    def _iter_pkg_manifests_from_category(
        self,
        category_dir: str,
    ) -> Iterable[BoundPackageManifest]:
        self.ensure_git_repo()

        category = os.path.basename(category_dir)

        # all valid semver strings start with a number
        for f in glob.iglob("*/[0-9]*.toml", root_dir=category_dir):
            pkg_name, pkg_ver = os.path.split(f)
            pkg_ver = pkg_ver[:-5]  # strip the ".toml" suffix
            with open(os.path.join(category_dir, f), "rb") as fp:
                yield BoundPackageManifest(
                    category,
                    pkg_name,
                    pkg_ver,
                    cast(InputPackageManifestType, tomllib.load(fp)),
                    self,
                )

    def get_supported_arches(self) -> list[str]:
        if self._supported_arches is not None:
            return list(self._supported_arches)

        res: set[str] = set()
        for plugin_id in self.iter_plugin_ids():
            if plugin_id.startswith("ruyi-profile-"):
                res.add(plugin_id[13:])
        self._supported_arches = res
        return list(res)

    def get_profile(self, name: str) -> ProfileProxy | None:
        # TODO: deprecate this after making sure every call site has gained
        # arch-awareness
        for arch in self.get_supported_arches():
            store = self.ensure_profile_store_for_arch(arch)
            if p := store.get(name):
                return p
        return None

    def get_profile_for_arch(self, arch: str, name: str) -> ProfileProxy | None:
        store = self.ensure_profile_store_for_arch(arch)
        return store.get(name)

    def iter_profiles_for_arch(self, arch: str) -> Iterable[ProfileProxy]:
        store = self.ensure_profile_store_for_arch(arch)
        return store.iter_profiles()

    def ensure_profile_store_for_arch(self, arch: str) -> ArchProfileStore:
        if arch in self._arch_profile_stores:
            return self._arch_profile_stores[arch]

        self.ensure_git_repo()
        store = ArchProfileStore(self._plugin_host_ctx, arch)
        self._arch_profile_stores[arch] = store
        return store

    def ensure_pkg_cache(
        self,
        ensure_repo: bool = True,
    ) -> None:
        if self._pkgs:
            return

        if ensure_repo:
            self.ensure_git_repo()

        cache_by_name: dict[str, dict[str, BoundPackageManifest]] = {}
        cache_by_category: dict[str, dict[str, dict[str, BoundPackageManifest]]] = {}
        slug_cache: dict[str, BoundPackageManifest] = {}
        for pm in self.iter_pkg_manifests(ensure_repo=ensure_repo):
            if pm.name not in cache_by_name:
                cache_by_name[pm.name] = {}
            cache_by_name[pm.name][pm.ver] = pm

            if pm.category not in cache_by_category:
                cache_by_category[pm.category] = {pm.name: {}}
            if pm.name not in cache_by_category[pm.category]:
                cache_by_category[pm.category][pm.name] = {}
            cache_by_category[pm.category][pm.name][pm.ver] = pm

            if pm.slug:
                slug_cache[pm.slug] = pm

        self._pkgs = cache_by_name
        self._categories = cache_by_category
        self._slug_cache = slug_cache

    def iter_pkgs(
        self,
        ensure_repo: bool = True,
    ) -> Iterable[tuple[str, str, dict[str, BoundPackageManifest]]]:
        if not self._pkgs:
            self.ensure_pkg_cache(ensure_repo=ensure_repo)

        for cat, cat_pkgs in self._categories.items():
            for pkg_name, pkg_vers in cat_pkgs.items():
                yield (cat, pkg_name, pkg_vers)

    def get_pkg_by_slug(
        self,
        slug: str,
        ensure_repo: bool = True,
    ) -> BoundPackageManifest | None:
        if not self._pkgs:
            self.ensure_pkg_cache(ensure_repo=ensure_repo)

        return self._slug_cache.get(slug)

    def iter_pkg_vers(
        self,
        name: str,
        category: str | None = None,
        ensure_repo: bool = True,
    ) -> Iterable[BoundPackageManifest]:
        if not self._pkgs:
            self.ensure_pkg_cache(ensure_repo=ensure_repo)

        if category is not None:
            return self._categories[category][name].values()
        return self._pkgs[name].values()

    def get_pkg(
        self,
        name: str,
        category: str,
        ver: str,
        *,
        ensure_repo: bool = True,
    ) -> BoundPackageManifest | None:
        if not self._pkgs:
            self.ensure_pkg_cache(ensure_repo=ensure_repo)

        try:
            return self._categories[category][name][ver]
        except KeyError:
            return None

    def get_pkg_latest_ver(
        self,
        name: str,
        category: str | None = None,
        include_prerelease_vers: bool = False,
        ensure_repo: bool = True,
    ) -> BoundPackageManifest:
        if not self._pkgs:
            self.ensure_pkg_cache(ensure_repo=ensure_repo)

        if category is not None:
            pkgset = self._categories[category]
        else:
            pkgset = self._pkgs

        all_semvers = [pm.semver for pm in pkgset[name].values()]
        if not include_prerelease_vers:
            all_semvers = [sv for sv in all_semvers if not is_prerelease(sv)]
        latest_ver = max(all_semvers)
        return pkgset[name][str(latest_ver)]

    def get_distfile_urls(self, decl: DistfileDecl) -> list[str]:
        urls_to_expand: list[str] = []
        if not decl.is_restricted("mirror"):
            urls_to_expand.append(f"mirror://{MIRROR_ID_RUYI_DIST}/{decl.name}")

        if decl.urls:
            urls_to_expand.extend(decl.urls)

        cfg = self.config
        return list(
            itertools.chain(
                *(
                    cfg.get_dist_urls_for_file(self.logger, url)
                    for url in urls_to_expand
                )
            )
        )

    def _ensure_news_cache(
        self,
        ensure_repo: bool = True,
    ) -> None:
        if self._news_cache is not None:
            return

        if ensure_repo:
            self.ensure_git_repo()
        news_dir = os.path.join(self.root, "news")

        rs_store = self._gc.news_read_status
        rs_store.load()

        cache = NewsItemStore(rs_store)
        try:
            for f in glob.iglob("*.md", root_dir=news_dir):
                with open(os.path.join(news_dir, f), "r", encoding="utf-8") as fp:
                    try:
                        contents = fp.read()
                    except UnicodeDecodeError:
                        self.logger.W(
                            _("UnicodeDecodeError: {path}").format(
                                path=os.path.join(news_dir, f)
                            )
                        )
                        continue
                    cache.add(f, contents)  # may fail but failures are harmless
        except FileNotFoundError:
            pass

        cache.finalize()
        self._news_cache = cache

    def news_store(
        self,
        ensure_repo: bool = True,
    ) -> NewsItemStore:
        if self._news_cache is None:
            self._ensure_news_cache(ensure_repo=ensure_repo)
        assert self._news_cache is not None
        return self._news_cache

    def run_plugin_cmd(self, cmd_name: str, args: list[str]) -> int:
        plugin_id = f"ruyi-cmd-{cmd_name.lower()}"

        plugin_entrypoint = self._plugin_host_ctx.get_from_plugin(
            plugin_id,
            "plugin_cmd_main_v1",
            is_cmd_plugin=True,  # allow access to host FS for command plugins
        )
        if plugin_entrypoint is None:
            raise RuntimeError(f"cmd entrypoint not found in plugin '{plugin_id}'")

        ret = self.eval_plugin_fn(plugin_entrypoint, args)
        if not isinstance(ret, int):
            self.logger.W(
                _(
                    "unexpected return type of cmd plugin '{plugin_id}': {type} is not int."
                ).format(
                    plugin_id=plugin_id,
                    type=type(ret),
                )
            )
            self.logger.I(_("forcing return code to 1; the plugin should be fixed"))
            ret = 1
        return ret

    @cached_property
    def entity_store(self) -> EntityStore:
        """Get the entity store for this repository."""
        return EntityStore(
            self.logger,
            FSEntityProvider(self.logger, pathlib.Path(self.root) / "entities"),
            MetadataRepoEntityProvider(self),
            ProfileEntityProvider(self),
        )

    def get_telemetry_api_url(self, scope: TelemetryScopeConfig) -> str | None:
        # do not clone the metadata repo if it is absent, in case the user
        # is simply trying trivial commands like `ruyi version`.
        if repo_cfg := self.maybe_config:
            return repo_cfg.get_telemetry_api_url(scope)
        return None


PACKAGE_ENTITY_TYPE = "pkg"
PACKAGE_ENTITY_TYPE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "required": ["pkg"],
    "properties": {
        "pkg": {
            "type": "object",
            "properties": {
                "id": {"type": "string"},
                "display_name": {"type": "string"},
                "name": {"type": "string"},
                "category": {"type": "string"},
            },
            "required": ["id", "display_name", "name", "category"],
        },
        "related": {
            "type": "array",
            "description": "List of related entity references",
            "items": {"type": "string", "pattern": "^.+:.+"},
        },
        "unique_among_type_during_traversal": {
            "type": "boolean",
            "description": "Whether this entity should be unique among all entities of the same type during traversal",
        },
    },
}


class PackageEntityData(TypedDict):
    id: str
    display_name: str
    name: str
    category: str


class PackageEntity(TypedDict):
    pkg: PackageEntityData
    related: "NotRequired[list[str]]"
    unique_among_type_during_traversal: "NotRequired[bool]"


class MetadataRepoEntityProvider(BaseEntityProvider):
    def __init__(self, repo: MetadataRepo) -> None:
        super().__init__()
        self._repo = repo

    def discover_schemas(self) -> dict[str, object]:
        return {
            PACKAGE_ENTITY_TYPE: PACKAGE_ENTITY_TYPE_SCHEMA,
        }

    def load_entities(
        self,
        entity_types: Sequence[str],
    ) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
        result: dict[str, Mapping[str, Mapping[str, Any]]] = {}
        for ty in entity_types:
            if ty == PACKAGE_ENTITY_TYPE:
                result[ty] = self._load_package_entities()
        return result

    def _load_package_entities(self) -> dict[str, PackageEntity]:
        result: dict[str, PackageEntity] = {}
        for cat, pkg_name, pkg_vers in self._repo.iter_pkgs():
            full_name = f"{cat}/{pkg_name}"
            relations = []

            # see if all versions of the package are toolchains and share the
            # same arch
            tc_arch: str | None = None
            for pkg_ver in pkg_vers.values():
                if tm := pkg_ver.toolchain_metadata:
                    if tc_arch is None:
                        tc_arch = tm.target_arch
                        continue
                    if tc_arch != tm.target_arch:
                        tc_arch = None
                        break
                else:
                    break
            if tc_arch is not None:
                # this is a toolchain package, add the arch as a related entity
                relations.append(f"arch:{tc_arch}")

            # similarly, check for the emulator kind
            emu_arches: set[str] | None = None
            for pkg_ver in pkg_vers.values():
                if em := pkg_ver.emulator_metadata:
                    pkg_ver_arches: set[str] = set()
                    for p in em.programs:
                        pkg_ver_arches.update(p.supported_arches)
                    if emu_arches is None:
                        emu_arches = pkg_ver_arches
                        continue
                    if emu_arches != pkg_ver_arches:
                        emu_arches = emu_arches.intersection(pkg_ver_arches)
                else:
                    break
            if emu_arches is not None:
                for emu_arch in emu_arches:
                    relations.append(f"arch:{emu_arch}")

            result[full_name] = {
                "pkg": {
                    "id": full_name,
                    "display_name": full_name,
                    "name": pkg_name,
                    "category": cat,
                },
                "related": relations,
            }
        return result


class PorcelainRepoEntryV1(PorcelainEntity):
    id: str
    name: str
    remote: str | None
    branch: str
    local_path: str | None
    priority: int
    active: bool
    is_system: bool



from rich import box
from rich.table import Table

from ..config import GlobalConfig
from ..i18n import _
from ..log import RuyiLogger
from ..utils.markdown import RuyiStyledMarkdown
from ..utils.porcelain import PorcelainOutput
from .news_store import NewsItem, NewsItemContent, NewsItemStore


def print_news_item_titles(
    logger: RuyiLogger,
    newsitems: list[NewsItem],
    lang: str,
) -> None:
    tbl = Table(box=box.SIMPLE, show_edge=False)
    # i18n NOTE: used as news item table title
    tbl.add_column(_("No."))
    # i18n NOTE: used as news item table title
    tbl.add_column(_("ID"))
    # i18n NOTE: used as news item table title
    tbl.add_column(_("Title"))

    for ni in newsitems:
        unread = not ni.is_read
        ord = f"[bold green]{ni.ordinal}[/]" if unread else f"{ni.ordinal}"
        id = f"[bold green]{ni.id}[/]" if unread else ni.id

        tbl.add_row(
            ord,
            id,
            ni.get_content_for_lang(lang).display_title,
        )

    logger.stdout(tbl)


def maybe_notify_unread_news(
    gc: GlobalConfig,
    prompt_no_unread: bool = True,
) -> None:
    """Check if there are new newsitems, notify the user if so."""

    unread_newsitems = gc.repo.news_store().list(True)
    if unread_newsitems:
        gc.logger.stdout(
            _("\nThere are {count} new news item(s):\n").format(
                count=len(unread_newsitems),
            )
        )
        print_news_item_titles(gc.logger, unread_newsitems, gc.lang_code)
        gc.logger.stdout(_("\nYou can read them with [yellow]ruyi news read[/]."))
        return

    if prompt_no_unread:
        gc.logger.stdout(
            _(
                "\nAll news items have been read. To see a list of them, run [yellow]ruyi news list[/].\n"
            )
        )


def do_news_list(
    cfg: GlobalConfig,
    only_unread: bool,
) -> int:
    logger = cfg.logger
    store = cfg.repo.news_store()
    newsitems = store.list(only_unread)

    if cfg.is_porcelain:
        with PorcelainOutput() as po:
            for ni in newsitems:
                po.emit(ni.to_porcelain())
        return 0

    logger.stdout(_("[bold green]News items:[/]\n"))
    if not newsitems:
        logger.stdout(_("  (no unread item)") if only_unread else _("  (no item)"))
        return 0

    print_news_item_titles(logger, newsitems, cfg.lang_code)

    return 0


def do_news_read(
    cfg: GlobalConfig,
    quiet: bool,
    items_strs: list[str],
) -> int:
    logger = cfg.logger
    store = cfg.repo.news_store()

    # filter out requested news items
    items = filter_news_items_by_specs(logger, store, items_strs)
    if items is None:
        return 1

    if cfg.is_porcelain:
        with PorcelainOutput() as po:
            for ni in items:
                po.emit(ni.to_porcelain())
    elif not quiet:
        # render the items
        if items:
            for ni in items:
                print_news(logger, ni.get_content_for_lang(cfg.lang_code))
        else:
            logger.stdout(_("No news to display."))

    # record read statuses
    store.mark_as_read(*(ni.id for ni in items))

    return 0


def filter_news_items_by_specs(
    logger: RuyiLogger,
    store: NewsItemStore,
    specs: list[str],
) -> list[NewsItem] | None:
    if not specs:
        # all unread items
        return store.list(True)

    all_ni = store.list(False)
    items: list[NewsItem] = []
    ni_by_ord = {ni.ordinal: ni for ni in all_ni}
    ni_by_id = {ni.id: ni for ni in all_ni}
    for i in specs:
        try:
            ni_ord = int(i)
            if ni_ord not in ni_by_ord:
                logger.F(
                    _("there is no news item with ordinal {ord}").format(ord=ni_ord)
                )
                return None
            items.append(ni_by_ord[ni_ord])
        except ValueError:
            # treat i as id
            if i not in ni_by_id:
                logger.F(_("there is no news item with ID '{id}'").format(id=i))
                return None
            items.append(ni_by_id[i])

    return items


def print_news(logger: RuyiLogger, nic: NewsItemContent) -> None:
    md = RuyiStyledMarkdown(nic.content)
    logger.stdout(md)
    logger.stdout("")



import argparse
import pathlib
from typing import TYPE_CHECKING

from ...cli.cmd import RootCommand
from ...i18n import _

if TYPE_CHECKING:
    from ...cli.completion import ArgumentParser
    from ...config import GlobalConfig


class VenvCommand(
    RootCommand,
    cmd="venv",
    help=_("Generate a virtual environment adapted to the chosen toolchain and profile"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.formatter_class = argparse.RawDescriptionHelpFormatter
        p.epilog = _(
            "Sysroot provisioning:\n"
            "  By default, Ruyi uses the sysroot bundled with the selected toolchain if one is available.\n"
            "  Use --copy-sysroot-from-pkg to copy the sysroot from another installed toolchain package.\n"
            "  Use --copy-sysroot-from-dir only for a complete sysroot directory readable by the current user; it performs a faithful full-tree copy.\n"
            "  Use --symlink-sysroot-from-dir to point the virtual environment at an existing sysroot directory without copying it.\n"
            "  Use --project-sysroot-from-rootfs for distro rootfs or chroot trees: Ruyi copies common cross-build directories such as include, lib*, usr/include, usr/lib*, usr/share, bin, and sbin, and skips unreadable or unsupported files.\n"
            "  Ruyi never elevates privileges when creating virtual environments. If a rootfs contains private system files such as /etc/shadow, prepare a readable sysroot yourself or use projection mode."
        )
        p.add_argument("profile", type=str, help=_("Profile to use for the environment"),
                       )
        p.add_argument("dest", type=str, help=_("Path to the new virtual environment"),
                       )
        p.add_argument(
            "--name",
            "-n",
            type=str,
            default=None,
            help=_("Override the venv's name"),
        )
        p.add_argument(
            "--toolchain",
            "-t",
            type=str,
            action="append",
            help=_("Specifier(s) (atoms) of the toolchain package(s) to use"),
        )
        p.add_argument(
            "--emulator",
            "-e",
            type=str,
            help=_("Specifier (atom) of the emulator package to use"),
        )
        p.add_argument(
            "--with-sysroot",
            action="store_true",
            dest="with_sysroot",
            default=True,
            help=_("Provision a fresh sysroot inside the new virtual environment (default)"),
        )
        p.add_argument(
            "--without-sysroot",
            action="store_false",
            dest="with_sysroot",
            help=_("Do not include a sysroot inside the new virtual environment"),
        )
        p.add_argument(
            "--copy-sysroot-from-pkg",
            "--sysroot-from",
            type=str,
            dest="copy_sysroot_from_pkg",
            help=_("Specifier (atom) of the sysroot package to use, in favor of the toolchain-included one if applicable"),
        )
        p.add_argument(
            "--copy-sysroot-from-dir",
            type=str,
            help=_("Copy the sysroot from the given directory into the virtual environment"),
        )
        p.add_argument(
            "--symlink-sysroot-from-dir",
            type=str,
            help=_("Symlink the virtual environment's sysroot to the given existing directory"),
        )
        p.add_argument(
            "--project-sysroot-from-rootfs",
            type=str,
            help=_("Project a build sysroot from the given distro rootfs directory"),
        )
        p.add_argument(
            "--extra-commands-from",
            type=str,
            action="append",
            help=_("Specifier(s) (atoms) of extra package(s) to add commands to the new virtual environment"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from ...ruyipkg.host import get_native_host
        from .maker import do_make_venv

        # validate sysroot source options: at most one source may be specified
        sysroot_sources = sum([
            args.copy_sysroot_from_pkg is not None,
            args.copy_sysroot_from_dir is not None,
            args.symlink_sysroot_from_dir is not None,
            args.project_sysroot_from_rootfs is not None,
        ])
        if sysroot_sources > 1:
            cfg.logger.F(
                _("at most one of --copy-sysroot-from-pkg, --copy-sysroot-from-dir, --symlink-sysroot-from-dir, and --project-sysroot-from-rootfs may be specified")
            )
            return 1

        if not args.with_sysroot and sysroot_sources > 0:
            cfg.logger.F(
                _("--without-sysroot cannot be combined with a sysroot source option")
            )
            return 1

        profile_name: str = args.profile
        dest = pathlib.Path(args.dest)
        with_sysroot: bool = args.with_sysroot
        override_name: str | None = args.name
        tc_atoms_str: list[str] | None = args.toolchain
        emu_atom_str: str | None = args.emulator
        sysroot_atom_str: str | None = args.copy_sysroot_from_pkg
        copy_sysroot_dir_str: str | None = args.copy_sysroot_from_dir
        symlink_sysroot_dir_str: str | None = args.symlink_sysroot_from_dir
        project_sysroot_dir_str: str | None = args.project_sysroot_from_rootfs
        extra_cmd_atoms_str: list[str] | None = args.extra_commands_from
        host = str(get_native_host())

        return do_make_venv(
            cfg,
            host,
            profile_name,
            dest,
            with_sysroot,
            override_name,
            tc_atoms_str,
            emu_atom_str,
            sysroot_atom_str,
            copy_sysroot_dir_str,
            symlink_sysroot_dir_str,
            project_sysroot_dir_str,
            extra_cmd_atoms_str,
        )



from contextlib import redirect_stderr, redirect_stdout
import json
import pathlib

import pytest

from ruyi.cli.main import main as ruyi_main
from tests.fixtures import IntegrationTestHarness


SHA_STUB = "0" * 64


CANONICAL_MANIFEST = f"""format = "v1"

[metadata]
desc = "Test package"
vendor = {{ name = "Test Vendor", eula = "" }}

[[distfiles]]
name = "src.tar.zst"
size = 0

[distfiles.checksums]
sha256 = "{SHA_STUB}"

[source]
distfiles = ["src.tar.zst"]
"""


NON_CANONICAL_MANIFEST = f"""format = "v1"

[[distfiles]]
size = 0
name = "src.tar.zst"

[distfiles.checksums]
sha256 = "{SHA_STUB}"

[metadata]
vendor = {{ eula = "", name = "Test Vendor" }}
desc = "Test package"

[source]
distfiles = ["src.tar.zst"]
"""


def test_admin_check_file_exits_zero_for_good_manifest(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(CANONICAL_MANIFEST, encoding="utf-8")

    result = ruyi_cli_runner("admin", "check", "-f", str(manifest_path))

    assert result.exit_code == 0
    assert "0 error(s), 0 warning(s)" in result.stdout


def test_admin_check_accepts_multiple_file_flags(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    first_path = tmp_path / "1.0.0.toml"
    second_path = tmp_path / "2.0.0.toml"
    first_path.write_text(CANONICAL_MANIFEST, encoding="utf-8")
    second_path.write_text(CANONICAL_MANIFEST, encoding="utf-8")

    result = ruyi_cli_runner(
        "admin",
        "check",
        "-f",
        str(first_path),
        "--file",
        str(second_path),
    )

    assert result.exit_code == 0
    assert "0 error(s), 0 warning(s)" in result.stdout


def test_admin_check_file_reports_diagnostics(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(NON_CANONICAL_MANIFEST, encoding="utf-8")

    result = ruyi_cli_runner("admin", "check", "-f", str(manifest_path))

    assert result.exit_code == 1
    assert "RYC0001" in result.stdout
    assert str(manifest_path) in result.stdout


def test_admin_check_repo_exits_zero_for_default_fixture_repo(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    result = ruyi_cli_runner(
        "admin",
        "check",
        "--repo",
        str(ruyi_cli_runner.repo_root),
    )

    assert result.exit_code == 0
    assert "0 error(s), 0 warning(s)" in result.stdout


def test_admin_check_repo_reports_multiple_diagnostics(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    repo_root = ruyi_cli_runner.repo_root
    ruyi_cli_runner.add_package("source", "bad-toml", "1.0.0", 'format = "v1"\n[')
    ruyi_cli_runner.add_package(
        "source",
        "bad-version",
        "not-semver",
        CANONICAL_MANIFEST,
    )

    result = ruyi_cli_runner("admin", "check", "--repo", str(repo_root))

    assert result.exit_code == 1
    assert "RYC0002" in result.stdout
    assert "RYC0004" in result.stdout


def test_admin_check_file_and_repo_fail_argument_validation(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(CANONICAL_MANIFEST, encoding="utf-8")

    with pytest.raises(SystemExit):
        ruyi_cli_runner(
            "admin",
            "check",
            "-f",
            str(manifest_path),
            "--repo",
            str(ruyi_cli_runner.repo_root),
        )


def test_admin_check_help_is_registered(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    ctx = ruyi_cli_runner.make_command_context("admin", "check", "--help")

    with (
        pytest.raises(SystemExit) as exc_info,
        redirect_stdout(ctx.stdout),
        redirect_stderr(ctx.stderr),
    ):
        ruyi_main(ctx.gm, ctx.gc, ctx.argv)

    assert exc_info.value.code == 0
    assert "usage: ruyi admin check" in ctx.stdout.getvalue()
    assert "--repo" in ctx.stdout.getvalue()
    assert "--only-packages" in ctx.stdout.getvalue()


def test_admin_check_only_packages_uses_list_style_selectors(
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    source_path = ruyi_cli_runner.add_package(
        "source",
        "ignored-source",
        "1.0.0",
        'format = "v1"\n[',
    )
    board_path = ruyi_cli_runner.add_package(
        "board-image",
        "selected-board",
        "1.0.0",
        NON_CANONICAL_MANIFEST,
    )

    result = ruyi_cli_runner(
        "admin",
        "check",
        "--repo",
        str(ruyi_cli_runner.repo_root),
        "--only-packages",
        "--category-is",
        "board-image",
    )

    assert result.exit_code == 1
    assert "RYC0001" in result.stdout
    assert str(board_path) in result.stdout
    assert str(source_path) not in result.stdout
    assert "RYC0002" not in result.stdout


def test_admin_check_porcelain_emits_jsonl_diagnostics(
    tmp_path: pathlib.Path,
    ruyi_cli_runner: IntegrationTestHarness,
) -> None:
    manifest_path = tmp_path / "1.0.0.toml"
    manifest_path.write_text(NON_CANONICAL_MANIFEST, encoding="utf-8")

    result = ruyi_cli_runner(
        "--porcelain",
        "admin",
        "check",
        "-f",
        str(manifest_path),
    )

    assert result.exit_code == 1
    entities = [json.loads(line) for line in result.stdout.splitlines()]
    assert [entity["ty"] for entity in entities] == ["checkdiagnostic-v1"]
    assert entities[0]["code"] == "RYC0001"



import mmap
import os
import shutil
import subprocess
from typing import Any, BinaryIO, NoReturn, Protocol

from ..i18n import _
from ..log import RuyiLogger
from ..utils import ar, prereqs
from .unpack_method import (
    UnpackMethod,
    UnrecognizedPackFormatError,
    determine_unpack_method,
)


class SupportsRead(Protocol):
    def read(self, n: int = -1, /) -> bytes: ...


def do_unpack(
    logger: RuyiLogger,
    filename: str,
    dest: str | os.PathLike[Any] | None,
    strip_components: int,
    unpack_method: UnpackMethod,
    stream: BinaryIO | SupportsRead | None = None,
    prefixes_to_unpack: list[str] | None = None,
) -> None:
    match unpack_method:
        case UnpackMethod.AUTO:
            raise ValueError("the auto unpack method must be resolved prior to use")
        case UnpackMethod.RAW:
            return _do_copy_raw(filename, dest)
        case (
            UnpackMethod.TAR_AUTO
            | UnpackMethod.TAR
            | UnpackMethod.TAR_BZ2
            | UnpackMethod.TAR_GZ
            | UnpackMethod.TAR_LZ4
            | UnpackMethod.TAR_XZ
            | UnpackMethod.TAR_ZST
        ):
            return _do_unpack_tar(
                logger,
                filename,
                dest,
                strip_components,
                unpack_method,
                stream,
                prefixes_to_unpack,
            )
        case UnpackMethod.ZIP:
            # TODO: handle strip_components somehow; the unzip(1) command currently
            # does not have such support.
            return _do_unpack_zip(logger, filename, dest)
        case UnpackMethod.DEB:
            return _do_unpack_deb(logger, filename, dest)
        case UnpackMethod.GZ:
            # bare gzip file
            return _do_unpack_bare_gz(logger, filename, dest)
        case UnpackMethod.BZ2:
            # bare bzip2 file
            return _do_unpack_bare_bzip2(logger, filename, dest)
        case UnpackMethod.LZ4:
            # bare lz4 file
            return _do_unpack_bare_lz4(logger, filename, dest)
        case UnpackMethod.XZ:
            # bare xz file
            return _do_unpack_bare_xz(logger, filename, dest)
        case UnpackMethod.ZST:
            # bare zstd file
            return _do_unpack_bare_zstd(logger, filename, dest)
        case _:
            raise UnrecognizedPackFormatError(filename)


def do_unpack_or_symlink(
    logger: RuyiLogger,
    filename: str,
    dest: str | os.PathLike[Any] | None,
    strip_components: int,
    unpack_method: UnpackMethod,
    stream: BinaryIO | SupportsRead | None = None,
    prefixes_to_unpack: list[str] | None = None,
) -> None:
    try:
        return do_unpack(
            logger,
            filename,
            dest,
            strip_components,
            unpack_method,
            stream,
            prefixes_to_unpack,
        )
    except UnrecognizedPackFormatError:
        # just symlink into destination
        return do_symlink(filename, dest)


def _do_copy_raw(
    src_path: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    src_filename = os.path.basename(src_path)
    if destdir is None:
        # symlink into CWD
        dest = src_filename
    else:
        dest = os.path.join(destdir, src_filename)

    shutil.copy(src_path, dest)


def do_symlink(
    src_path: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    src_filename = os.path.basename(src_path)
    if destdir is None:
        # symlink into CWD
        dest = src_filename
    else:
        dest = os.path.join(destdir, src_filename)

    # avoid the hassle and pitfalls around relative paths and symlinks, and
    # just point to the target using absolute path
    symlink_target = os.path.abspath(src_path)
    os.symlink(symlink_target, dest)


def _do_unpack_tar(
    logger: RuyiLogger,
    filename: str,
    dest: str | os.PathLike[Any] | None,
    strip_components: int,
    unpack_method: UnpackMethod,
    stream: SupportsRead | None,
    prefixes_to_unpack: list[str] | None = None,
) -> None:
    argv = ["tar", "-x"]

    match unpack_method:
        case UnpackMethod.TAR | UnpackMethod.TAR_AUTO:
            pass
        case UnpackMethod.TAR_GZ:
            argv.append("-z")
        case UnpackMethod.TAR_BZ2:
            argv.append("-j")
        case UnpackMethod.TAR_LZ4:
            argv.append("--use-compress-program=lz4")
        case UnpackMethod.TAR_XZ:
            argv.append("-J")
        case UnpackMethod.TAR_ZST:
            argv.append("--zstd")
        case _:
            raise ValueError(
                f"do_unpack_tar cannot handle non-tar unpack method {unpack_method}"
            )

    stdin: int | None = None
    if stream is not None:
        filename = "-"
        stdin = subprocess.PIPE

    argv.extend(("-f", filename, f"--strip-components={strip_components}"))
    if prefixes_to_unpack:
        if any(p.startswith("-") for p in prefixes_to_unpack):
            raise ValueError(
                "prefixes_to_unpack must not contain any item starting with '-'"
            )
        argv.extend(prefixes_to_unpack)
    logger.D(f"about to call tar: argv={argv}")
    p = subprocess.Popen(argv, cwd=dest, stdin=stdin)

    retcode: int
    if stream is None:
        retcode = p.wait()
    else:
        # this is only for pleasing the type-checker; it's statically true
        # because the assignment always happens due to the earlier
        # "stream is not None" branch.
        assert p.stdin is not None

        bufsize = 4 * mmap.PAGESIZE
        while True:
            buf = stream.read(bufsize)
            if not buf:
                break
            p.stdin.write(buf)
        p.stdin.close()
        retcode = p.wait()

    if retcode != 0:
        raise RuntimeError(
            _("untar failed: command {cmd} returned {retcode}").format(
                cmd=" ".join(argv),
                retcode=retcode,
            )
        )


def _do_unpack_zip(
    logger: RuyiLogger,
    filename: str,
    dest: str | os.PathLike[Any] | None,
) -> None:
    argv = ["unzip", filename]
    if dest is not None:
        argv.extend(("-d", str(dest)))
    logger.D(f"about to call unzip: argv={argv}")
    retcode = subprocess.call(argv, cwd=dest)
    if retcode != 0:
        raise RuntimeError(
            _("unzip failed: command {cmd} returned {retcode}").format(
                cmd=" ".join(argv),
                retcode=retcode,
            )
        )


def _do_unpack_bare_gz(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    # the suffix may not be ".gz" so do this generically
    dest_filename = os.path.splitext(os.path.basename(filename))[0]

    argv = ["gunzip", "-c", filename]
    if destdir is not None:
        os.chdir(destdir)

    logger.D(f"about to call gunzip: argv={argv}")
    with open(dest_filename, "wb") as out:
        retcode = subprocess.call(argv, stdout=out)
        if retcode != 0:
            raise RuntimeError(
                _("gunzip failed: command {cmd} returned {retcode}").format(
                    cmd=" ".join(argv),
                    retcode=retcode,
                )
            )


def _do_unpack_bare_bzip2(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    # the suffix may not be ".bz2" so do this generically
    dest_filename = os.path.splitext(os.path.basename(filename))[0]

    argv = ["bzip2", "-dc", filename]
    if destdir is not None:
        os.chdir(destdir)

    logger.D(f"about to call bzip2: argv={argv}")
    with open(dest_filename, "wb") as out:
        retcode = subprocess.call(argv, stdout=out)
        if retcode != 0:
            raise RuntimeError(
                _("bzip2 failed: command {cmd} returned {retcode}").format(
                    cmd=" ".join(argv),
                    retcode=retcode,
                )
            )


def _do_unpack_bare_lz4(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    # the suffix may not be ".lz4" so do this generically
    dest_filename = os.path.splitext(os.path.basename(filename))[0]

    argv = ["lz4", "-dk", filename, f"./{dest_filename}"]
    logger.D(f"about to call lz4: argv={argv}")
    retcode = subprocess.call(argv, cwd=destdir)
    if retcode != 0:
        raise RuntimeError(
            _("lz4 failed: command {cmd} returned {retcode}").format(
                cmd=" ".join(argv),
                retcode=retcode,
            )
        )


def _do_unpack_bare_xz(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    # the suffix may not be ".xz" so do this generically
    dest_filename = os.path.splitext(os.path.basename(filename))[0]

    argv = ["xz", "-d", "-c", filename]
    if destdir is not None:
        os.chdir(destdir)

    logger.D(f"about to call xz: argv={argv}")
    with open(dest_filename, "wb") as out:
        retcode = subprocess.call(argv, stdout=out)
        if retcode != 0:
            raise RuntimeError(
                _("xz failed: command {cmd} returned {retcode}").format(
                    cmd=" ".join(argv),
                    retcode=retcode,
                )
            )


def _do_unpack_bare_zstd(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    # the suffix may not be ".zst" so do this generically
    dest_filename = os.path.splitext(os.path.basename(filename))[0]

    argv = ["zstd", "-d", filename, "-o", f"./{dest_filename}"]
    logger.D(f"about to call zstd: argv={argv}")
    retcode = subprocess.call(argv, cwd=destdir)
    if retcode != 0:
        raise RuntimeError(
            _("zstd failed: command {cmd} returned {retcode}").format(
                cmd=" ".join(argv),
                retcode=retcode,
            )
        )


def _do_unpack_deb(
    logger: RuyiLogger,
    filename: str,
    destdir: str | os.PathLike[Any] | None,
) -> None:
    with ar.ArpyArchiveWrapper(filename) as a:
        for f in a.infolist():
            name = f.name.decode("utf-8")
            if name.startswith("data.tar"):
                inner_unpack_method = determine_unpack_method(name)
                return _do_unpack_tar(
                    logger,
                    name,
                    destdir,
                    0,
                    inner_unpack_method,
                    a.open(f),
                )

    raise RuntimeError(
        _("file '{filename}' does not appear to be a deb").format(
            filename=filename,
        )
    )


def _get_unpack_cmds_for_method(m: UnpackMethod) -> list[str]:
    match m:
        case UnpackMethod.UNKNOWN | UnpackMethod.RAW | UnpackMethod.DEB:
            return []
        case UnpackMethod.GZ:
            return ["gunzip"]
        case UnpackMethod.BZ2:
            return ["bzip2"]
        case UnpackMethod.LZ4:
            return ["lz4"]
        case UnpackMethod.XZ:
            return ["xz"]
        case UnpackMethod.ZST:
            return ["zstd"]
        case UnpackMethod.TAR | UnpackMethod.TAR_AUTO:
            return ["tar"]
        case UnpackMethod.TAR_GZ:
            return ["tar", "gunzip"]
        case UnpackMethod.TAR_BZ2:
            return ["tar", "bzip2"]
        case UnpackMethod.TAR_LZ4:
            return ["tar", "lz4"]
        case UnpackMethod.TAR_XZ:
            return ["tar", "xz"]
        case UnpackMethod.TAR_ZST:
            return ["tar", "zstd"]
        case UnpackMethod.ZIP:
            return ["unzip"]
        case UnpackMethod.AUTO:
            raise ValueError(f"the unpack method {m} must be resolved prior to use")


def ensure_unpack_cmd_for_method(
    logger: RuyiLogger,
    m: UnpackMethod,
) -> None | NoReturn:
    required_cmds = _get_unpack_cmds_for_method(m)
    if not required_cmds:
        return None

    return prereqs.ensure_cmds(logger, required_cmds, interactive_retry=True)



import ast

import pytest

from ruyi.pluginhost.unsandboxed import lint_module


def _lint(src: str) -> None:
    lint_module(ast.parse(src))


def test_plain_module_passes() -> None:
    _lint(
        "x = 1\n"
        "def f(a, b):\n"
        "    return a + b\n"
        "y = [f(i, 1) for i in range(10)]\n"
    )


def test_slice_in_value_position_passes() -> None:
    # Slicing on the RHS of an assignment, and as a subexpression on the
    # LHS but not as the outermost Subscript's slice, is legal Starlark.
    _lint("a = xs[1:3]\n")
    _lint("xs[ys[1:3]] = 0\n")


def test_starred_in_call_and_rhs_passes() -> None:
    # ``*args`` / ``**kwargs`` in call arguments and starred elements on
    # the RHS of an assignment are valid Starlark and must not be gated.
    _lint("f(*args, **kwargs)\n")
    _lint("xs = [*a, *b]\n")


@pytest.mark.parametrize(
    ("src", "expected_feature"),
    [
        ("y = (x := 1)\n", "walrus"),
        ("raise ValueError('x')\n", "raise"),
        ("assert True\n", "assert"),
        ("import os\n", "import"),
        ("from os import path\n", "from"),
        ("try:\n    pass\nexcept Exception:\n    pass\n", "try"),
        ("with open('x') as f:\n    pass\n", "with"),
        (
            "match x:\n    case 1:\n        pass\n    case _:\n        pass\n",
            "match",
        ),
        ("def g():\n    yield 1\n", "yield"),
        ("def g():\n    yield from [1]\n", "yield from"),
        ("def g():\n    global x\n", "global"),
        (
            "def outer():\n"
            "    x = 1\n"
            "    def inner():\n"
            "        nonlocal x\n"
            "    return inner\n",
            "nonlocal",
        ),
        ("class C:\n    pass\n", "class"),
        ("async def g():\n    pass\n", "async def"),
        ("@staticmethod\ndef f():\n    pass\n", "decorator"),
        ("x: int = 1\n", "variable type annotation"),
        ("def f() -> int:\n    return 1\n", "return type annotation"),
        ("def f(x: int):\n    return x\n", "parameter type annotation"),
        ("x = f'hello {name}'\n", "f-string"),
        ("x = {1, 2, 3}\n", "set display"),
        ("x = {i for i in range(3)}\n", "set comprehension"),
        ("x = sum(i for i in range(3))\n", "generator expression"),
        ("x = [1]\ndel x\n", "del"),
        ("x = a @ b\n", "matrix-multiplication operator"),
        ("a @= b\n", "matrix-multiplication assignment"),
        ("ok = 0 <= i < n\n", "chained comparison"),
        ("ok = a is b\n", "`is` operator"),
        ("ok = a is not b\n", "`is not` operator"),
        ("def f(a, /, b):\n    return a + b\n", "positional-only parameter"),
        ("def f():\n    while True:\n        pass\n", "while"),
        ("xs[1:3] = [0]\n", "slice as assignment target"),
        ("xs[1:3] += [0]\n", "slice as assignment target"),
        ("(xs[1:3], y[0]) = (0, 1)\n", "slice as assignment target"),
        ("[xs[1:3], y[0]] = [0, 1]\n", "slice as assignment target"),
        ("a, *rest = xs\n", "starred assignment target"),
        ("for *a, b in xs:\n    pass\n", "starred loop variable"),
        ("y = [x for *a, b in xs]\n", "starred loop variable"),
        ("y = {k: v for *a, (k, v) in xs}\n", "starred loop variable"),
        # `await`, `async for`, `async with` can only occur syntactically
        # inside an `async def`, so they are shadowed by the `async def`
        # rejection above; they have their own ``visit_*`` overrides anyway
        # for defence in depth.
    ],
)
def test_gated_feature_is_rejected(src: str, expected_feature: str) -> None:
    with pytest.raises(RuntimeError) as excinfo:
        _lint(src)
    msg = str(excinfo.value)
    assert "is not allowed in plugin code" in msg
    assert expected_feature in msg
    assert "line " in msg


@pytest.mark.parametrize(
    ("src", "expected_line", "expected_feature"),
    [
        # Walrus on the third line of the module.
        ("x = 1\ny = 2\nz = (w := 3)\n", 3, "walrus"),
        # ``del`` on line 2.
        ("x = [1]\ndel x\n", 2, "del"),
        # Decorator declared on line 1, its function header on line 2;
        # the gate reports the decorator's own line.
        ("@staticmethod\ndef f():\n    return 1\n", 1, "decorator"),
        # Nested gated construct: the inner ``while`` is on line 3.
        ("def f():\n    x = 1\n    while x:\n        x = 0\n", 3, "while"),
    ],
)
def test_gated_feature_reports_correct_line(
    src: str, expected_line: int, expected_feature: str
) -> None:
    with pytest.raises(RuntimeError) as excinfo:
        _lint(src)
    msg = str(excinfo.value)
    assert f"line {expected_line}:" in msg
    assert expected_feature in msg



from ..device import provision_cli as provision_cli
from ..mux.venv import venv_cli as venv_cli
from ..pluginhost import plugin_cli as plugin_cli
from ..ruyipkg import admin_cli as admin_cli
from ..ruyipkg import entity_cli as entity_cli
from ..ruyipkg import install_cli as install_cli
from ..ruyipkg import list_cli as list_cli
from ..ruyipkg import news_cli as news_cli
from ..ruyipkg import profile_cli as profile_cli
from ..ruyipkg import repo_cli as repo_cli
from ..ruyipkg import update_cli as update_cli
from ..telemetry import telemetry_cli as telemetry_cli
from . import self_cli as self_cli
from . import config_cli as config_cli
from . import version_cli as version_cli



#!/usr/bin/env python3
# Regenerates data.py from fresh contents.

import base64
import pathlib
from typing import Any
import zlib


def make_payload_from_file(path: pathlib.Path) -> str:
    with open(path, "rb") as fp:
        content = fp.read()

    return base64.b64encode(zlib.compress(content, 9)).decode("ascii")


def main() -> None:
    self_path = pathlib.Path(__file__).parent.resolve()
    bundled_resource_root = self_path / ".." / ".." / "resources" / "bundled"

    resources: dict[str, str] = {}
    template_names: dict[str, str] = {}
    for f in bundled_resource_root.glob("**/*"):
        if not f.is_file():
            continue

        rel_path = f.relative_to(bundled_resource_root)
        resources[str(rel_path)] = make_payload_from_file(f)

        if f.suffix.lower() == ".jinja":
            # strip the .jinja suffix for the template name
            template_names[str(rel_path.with_suffix(""))] = str(rel_path)

    with open(self_path / "data.py", "w", encoding="utf-8") as fp:

        def p(*args: Any) -> None:
            return print(*args, file=fp)

        p("# NOTE: This file is auto-generated. DO NOT EDIT!")
        p("# Update by running the __main__.py alongside this file\n")

        p("from typing import Final\n\n")

        p("RESOURCES: Final = {")
        for filename, payload in sorted(resources.items()):
            p(f'    "{filename}": b"{payload}",  # fmt: skip')
        p("}\n")

        p("TEMPLATE_NAME_MAP: Final = {")
        for stem, full_filename in sorted(template_names.items()):
            p(f'    "{stem}": "{full_filename}",')
        p("}")


if __name__ == "__main__":
    main()



"""
helper functions for CLI completions

see https://github.com/kislyuk/argcomplete/issues/443 for why this is needed
"""

import argparse
from typing import Any, Callable, Final, Optional, Sequence, cast

SUPPORTED_SHELLS: Final[set[str]] = {"bash", "zsh"}


class ArgcompleteAction(argparse.Action):
    completer: Optional[Callable[[str, object], list[str]]]

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        raise NotImplementedError(".__call__() not defined")


class ArgumentParser(argparse.ArgumentParser):
    def add_argument(self, *args: Any, **kwargs: Any) -> ArgcompleteAction:
        return cast(ArgcompleteAction, super().add_argument(*args, **kwargs))



import argparse
from typing import TYPE_CHECKING

from ..i18n import _
from .list_cli import ListCommand

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class ListProfilesCommand(
    ListCommand,
    cmd="profiles",
    help=_("List all available profiles"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        pass

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        logger = cfg.logger
        mr = cfg.repo

        for arch in mr.get_supported_arches():
            for p in mr.iter_profiles_for_arch(arch):
                if not p.need_quirks:
                    logger.stdout(
                        _("{profile_id} (arch: [green]{arch}[/])").format(
                            profile_id=p.id,
                            arch=arch,
                        )
                    )
                    continue

                logger.stdout(
                    _(
                        "{profile_id} (arch: [green]{arch}[/], needs quirks: [yellow]{need_quirks}[/])"
                    ).format(
                        profile_id=p.id,
                        arch=arch,
                        need_quirks=", ".join(sorted(p.need_quirks)),
                    )
                )

        return 0



from typing import Iterable, TypeAlias, TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import NotRequired

from ..utils.node_info import NodeInfo
from .event import TelemetryEvent


class AggregatedTelemetryEvent(TypedDict):
    time_bucket: str
    kind: str
    params: list[tuple[str, str]]
    count: int


class UploadPayload(TypedDict):
    fmt: int
    nonce: str
    ruyi_version: str
    report_uuid: "NotRequired[str]"
    """Optional field in case the client wishes to report this, and nothing
    else. If `installation` is present, this field is ignored."""
    installation: "NotRequired[NodeInfo | None]"
    """More detailed installation info that the client has user consent to report."""
    events: list[AggregatedTelemetryEvent]
    """Aggregated telemetry events that the client has user consent to upload."""


def stringify_param_val(v: object) -> str:
    if v is None:
        return "null"
    if isinstance(v, bool):
        return "1" if v else "0"
    if isinstance(v, bytes):
        return v.decode("utf-8")
    if isinstance(v, str):
        return v
    return str(v)


AggregateKey: TypeAlias = tuple[tuple[str, str], ...]


def _make_aggregate_key(ev: TelemetryEvent) -> AggregateKey:
    param_list = [(k, stringify_param_val(v)) for k, v in ev["params"].items()]
    param_list.sort()
    return tuple([("", ev["kind"])] + param_list)


def aggregate_events(
    events: Iterable[TelemetryEvent],
) -> Iterable[AggregatedTelemetryEvent]:
    # dict[time_bucket, dict[AggregateKey, count]]
    buf: dict[str, dict[AggregateKey, int]] = {}
    for raw_ev in events:
        time_bucket = raw_ev.get("time_bucket")
        if time_bucket is None:
            continue
        if time_bucket not in buf:
            buf[time_bucket] = {}

        agg_key = _make_aggregate_key(raw_ev)
        if agg_key not in buf[time_bucket]:
            buf[time_bucket][agg_key] = 1
        else:
            buf[time_bucket][agg_key] += 1

    for time_bucket in sorted(buf.keys()):
        bucket_events = buf[time_bucket]
        for agg_key in sorted(bucket_events.keys()):
            yield {
                "time_bucket": time_bucket,
                "kind": agg_key[0][1],
                "params": list(agg_key[1:]),
                "count": bucket_events[agg_key],
            }



from copy import deepcopy
from functools import cached_property
import os
import pathlib
import re
from typing import (
    Any,
    BinaryIO,
    Final,
    Iterable,
    Iterator,
    Literal,
    TypedDict,
    TYPE_CHECKING,
    cast,
)

if TYPE_CHECKING:
    from typing_extensions import NotRequired, Self

    # pyright only works with semver 3.x
    from semver.version import Version
else:
    try:
        from semver.version import Version  # type: ignore[import-untyped,unused-ignore]
    except ModuleNotFoundError:
        # semver 2.x
        from semver import VersionInfo as Version  # type: ignore[import-untyped,unused-ignore]

import tomlkit

from .host import RuyiHost, canonicalize_host_str, get_native_host
from .msg import RepoMessageStore
from .unpack_method import UnpackMethod, determine_unpack_method

if TYPE_CHECKING:
    # avoid circular import at runtime
    from .repo import MetadataRepo


class VendorDeclType(TypedDict):
    name: str
    eula: str | None


RestrictKind = Literal["fetch"] | Literal["mirror"]


class FetchRestrictionDeclType(TypedDict):
    msgid: str
    params: "NotRequired[dict[str, str]]"


class DistfileDeclType(TypedDict):
    name: str
    urls: "NotRequired[list[str]]"
    restrict: "NotRequired[list[RestrictKind]]"
    size: int
    checksums: dict[str, str]
    strip_components: "NotRequired[int]"
    unpack: "NotRequired[UnpackMethod]"
    fetch_restriction: "NotRequired[FetchRestrictionDeclType]"
    prefixes_to_unpack: "NotRequired[list[str]]"


class BinaryFileDeclType(TypedDict):
    host: str
    distfiles: list[str]
    commands: "NotRequired[dict[str, str]]"


BinaryDeclType = list[BinaryFileDeclType]


class BlobDeclType(TypedDict):
    distfiles: list[str]


class SourceDeclType(TypedDict):
    distfiles: list[str]


class ToolchainComponentDeclType(TypedDict):
    name: str
    version: str


class ToolchainDeclType(TypedDict):
    target: str
    quirks: "NotRequired[list[str]]"
    flavors: "NotRequired[list[str]]"  # Backward compatibility alias
    components: list[ToolchainComponentDeclType]
    included_sysroot: "NotRequired[str]"


EmulatorFlavor = Literal["qemu-linux-user"]


class EmulatorProgramDeclType(TypedDict):
    path: str
    flavor: EmulatorFlavor
    supported_arches: list[str]
    binfmt_misc: "NotRequired[str]"


class EmulatorDeclType(TypedDict):
    quirks: "NotRequired[list[str]]"
    flavors: "NotRequired[list[str]]"  # Backward compatibility alias
    programs: list[EmulatorProgramDeclType]


PartitionKind = (
    Literal["boot"]
    | Literal["disk"]
    | Literal["live"]
    | Literal["root"]
    | Literal["uboot"]
)

# error: "<typing special form>" has no attribute "__args__"
# KNOWN_PARTITION_KINDS = frozenset(kind.__args__[0] for kind in PartitionKind.__args__)
KNOWN_PARTITION_KINDS: Final = frozenset(("boot", "disk", "live", "root", "uboot"))

PartitionMapDecl = dict[PartitionKind, str]


class ProvisionableDeclType(TypedDict):
    partition_map: PartitionMapDecl
    strategy: str


PackageKind = (
    Literal["binary"]
    | Literal["blob"]
    | Literal["source"]
    | Literal["toolchain"]
    | Literal["emulator"]
    | Literal["provisionable"]
)

ALL_PACKAGE_KINDS: Final[list[PackageKind]] = [
    "binary",
    "blob",
    "source",
    "toolchain",
    "emulator",
    "provisionable",
]

RuyiPkgFormat = Literal["v1"]

ServiceLevelKind = Literal["known_issue"] | Literal["untested"]

ALL_SERVICE_LEVEL_KINDS: Final[list[ServiceLevelKind]] = ["known_issue", "untested"]


class ServiceLevelDeclType(TypedDict):
    level: ServiceLevelKind
    msgid: "NotRequired[str]"
    params: "NotRequired[dict[str, str]]"


class PackageMetadataDeclType(TypedDict):
    slug: "NotRequired[str]"  # deprecated for v1+
    desc: str
    doc_uri: "NotRequired[str]"
    vendor: VendorDeclType
    service_level: "NotRequired[list[ServiceLevelDeclType]]"
    upstream_version: "NotRequired[str]"


class InputPackageManifestType(TypedDict):
    format: "NotRequired[RuyiPkgFormat]"

    # v0 fields
    slug: "NotRequired[str]"
    kind: "NotRequired[list[PackageKind]]"  # mandatory in v0
    desc: "NotRequired[str]"  # mandatory in v0
    doc_uri: "NotRequired[str]"
    vendor: "NotRequired[VendorDeclType]"  # mandatory in v0

    # v1+ fields
    metadata: "NotRequired[PackageMetadataDeclType]"

    # common fields
    distfiles: list[DistfileDeclType]
    binary: "NotRequired[BinaryDeclType]"
    blob: "NotRequired[BlobDeclType]"
    source: "NotRequired[SourceDeclType]"
    toolchain: "NotRequired[ToolchainDeclType]"
    emulator: "NotRequired[EmulatorDeclType]"
    provisionable: "NotRequired[ProvisionableDeclType]"


class PackageManifestType(TypedDict):
    format: RuyiPkgFormat
    kind: list[PackageKind]
    metadata: PackageMetadataDeclType
    distfiles: list[DistfileDeclType]
    binary: "NotRequired[BinaryDeclType]"
    blob: "NotRequired[BlobDeclType]"
    source: "NotRequired[SourceDeclType]"
    toolchain: "NotRequired[ToolchainDeclType]"
    emulator: "NotRequired[EmulatorDeclType]"
    provisionable: "NotRequired[ProvisionableDeclType]"


class DistfileDecl:
    def __init__(self, data: DistfileDeclType) -> None:
        self._data = data

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def urls(self) -> list[str] | None:
        return self._data.get("urls")

    def is_restricted(self, kind: RestrictKind) -> bool:
        if restricts := self._data.get("restrict"):
            # account for a common oversight in some existing manifests where
            # the field was specified as a string instead of a list, in case the
            # user has not yet synced their repo
            if isinstance(restricts, str):
                return kind == restricts
            return kind in restricts
        return False

    @property
    def size(self) -> int:
        return self._data["size"]

    @property
    def checksums(self) -> dict[str, str]:
        return self._data["checksums"]

    def get_checksum(self, kind: str) -> str | None:
        return self._data["checksums"].get(kind)

    @property
    def prefixes_to_unpack(self) -> list[str] | None:
        return self._data.get("prefixes_to_unpack")

    @property
    def strip_components(self) -> int:
        return self._data.get("strip_components", 1)

    @property
    def unpack_method(self) -> UnpackMethod:
        x = self._data.get("unpack", UnpackMethod.AUTO)
        if x == UnpackMethod.AUTO:
            return determine_unpack_method(self.name)
        return x

    @property
    def fetch_restriction(self) -> FetchRestrictionDeclType | None:
        return self._data.get("fetch_restriction")


class BinaryDecl:
    def __init__(self, data: BinaryDeclType) -> None:
        self._data = {canonicalize_host_str(d["host"]): d for d in data}

    @property
    def data(self) -> dict[str, BinaryFileDeclType]:
        return self._data

    def get_distfile_names_for_host(self, host: str | RuyiHost) -> list[str] | None:
        if data := self._data.get(canonicalize_host_str(host)):
            return data.get("distfiles")
        return None

    @property
    def is_available_for_current_host(self) -> bool:
        return str(get_native_host()) in self._data

    def get_commands_for_host(self, host: str) -> dict[str, str]:
        if data := self._data.get(canonicalize_host_str(host)):
            return data.get("commands", {})
        return {}


class BlobDecl:
    def __init__(self, data: BlobDeclType) -> None:
        self._data = data

    def get_distfile_names(self) -> list[str] | None:
        return self._data["distfiles"]


class SourceDecl:
    def __init__(self, data: SourceDeclType) -> None:
        self._data = data

    def get_distfile_names_for_host(self, host: str | RuyiHost) -> list[str] | None:
        # currently the host parameter is ignored
        return self._data["distfiles"]


class ToolchainDecl:
    def __init__(self, data: ToolchainDeclType) -> None:
        self._data = data
        self._component_vers_cache: dict[str, str] | None = None

        # rename "flavors" to "quirks" for compatibility with old data
        if "quirks" not in self._data and "flavors" in self._data:
            self._data["quirks"] = self._data["flavors"]
            del self._data["flavors"]

    @property
    def _component_vers(self) -> dict[str, str]:
        if self._component_vers_cache is None:
            self._component_vers_cache = {
                x["name"]: x["version"] for x in self.components
            }
        return self._component_vers_cache

    @property
    def target(self) -> str:
        return self._data["target"]

    @property
    def target_arch(self) -> str:
        # TODO: switch to proper mapping later; for now this suffices
        return self.target.split("-", 1)[0]

    @property
    def quirks(self) -> list[str]:
        return self._data.get("quirks", [])

    def has_quirk(self, q: str) -> bool:
        return q in self.quirks

    def satisfies_quirk_set(self, req: set[str]) -> bool:
        # req - my_quirks must be the empty set so that my_quirks >= req
        return len(req.difference(self.quirks)) == 0

    @property
    def components(self) -> Iterable[ToolchainComponentDeclType]:
        return self._data["components"]

    def get_component_version(self, name: str) -> str | None:
        return self._component_vers.get(name)

    @property
    def has_binutils(self) -> bool:
        return self.get_component_version("binutils") is not None

    @property
    def has_clang(self) -> bool:
        return self.get_component_version("clang") is not None

    @property
    def has_gcc(self) -> bool:
        return self.get_component_version("gcc") is not None

    @property
    def has_llvm(self) -> bool:
        return self.get_component_version("llvm") is not None

    @property
    def included_sysroot(self) -> str | None:
        return self._data.get("included_sysroot")


class EmulatorProgDecl:
    def __init__(self, data: EmulatorProgramDeclType) -> None:
        self.relative_path = data["path"]
        # have to explicitly annotate the type to please the type checker...
        self.flavor: EmulatorFlavor = data["flavor"]
        self.supported_arches = set(data["supported_arches"])
        self.binfmt_misc = data.get("binfmt_misc")

    def get_binfmt_misc_str(self, install_root: os.PathLike[Any]) -> str | None:
        if self.binfmt_misc is None:
            return None
        binpath = os.path.join(install_root, self.relative_path)
        return self.binfmt_misc.replace("$BIN", binpath)

    @property
    def is_qemu(self) -> bool:
        return self.flavor == "qemu-linux-user"


class EmulatorDecl:
    def __init__(self, data: EmulatorDeclType) -> None:
        self._data = data
        self.programs = [EmulatorProgDecl(x) for x in data["programs"]]

        # rename "flavors" to "quirks" for compatibility with old data
        if "quirks" not in self._data and "flavors" in self._data:
            self._data["quirks"] = self._data["flavors"]
            del self._data["flavors"]

    @property
    def quirks(self) -> list[str] | None:
        return self._data.get("quirks")

    def list_for_arch(self, arch: str) -> Iterable[EmulatorProgDecl]:
        for p in self.programs:
            if arch in p.supported_arches:
                yield p


class ProvisionableDecl:
    def __init__(self, data: ProvisionableDeclType) -> None:
        self._data = data

    @property
    def partition_map(self) -> PartitionMapDecl:
        return self._data["partition_map"]

    @property
    def strategy(self) -> str:
        return self._data["strategy"]


class PackageMetadataDecl:
    def __init__(self, data: PackageMetadataDeclType) -> None:
        self._data = data


def _translate_to_manifest_v1(obj: InputPackageManifestType) -> PackageManifestType:
    fmt = obj.get("format", "")
    if fmt == "v1":
        return cast(PackageManifestType, obj)
    if fmt != "":
        # unrecognized package format
        raise RuntimeError(f"unrecognized Ruyi package format: {fmt}")

    # translate v0 to v1
    result = deepcopy(obj)
    result["format"] = "v1"

    md: PackageMetadataDeclType = {"desc": "", "vendor": {"name": "", "eula": None}}
    if "slug" in result:
        md["slug"] = result["slug"]
        del result["slug"]
    if "desc" in result:
        md["desc"] = result["desc"]
        del result["desc"]
    if "vendor" in result:
        md["vendor"] = result["vendor"]
        del result["vendor"]
    if "doc_uri" in result:
        md["doc_uri"] = result["doc_uri"]
        del result["doc_uri"]
    result["metadata"] = md

    return cast(PackageManifestType, result)


class PackageServiceLevel:
    def __init__(self, data: list[ServiceLevelDeclType]) -> None:
        self._data = data

    @property
    def level(self) -> ServiceLevelKind:
        for r in self._data:
            if r["level"] == "untested":
                continue
            return r["level"]
        return "untested"

    @property
    def has_known_issues(self) -> bool:
        for r in self._data:
            if r["level"] == "known_issue":
                return True
        return False

    @property
    def known_issues(self) -> Iterator[ServiceLevelDeclType]:
        for r in self._data:
            if r["level"] == "known_issue":
                yield r

    def render_known_issues(
        self,
        msg_store: RepoMessageStore,
        lang_code: str,
    ) -> Iterator[str]:
        for x in self.known_issues:
            if "msgid" not in x:
                # malformed known issue declaration, but let's not panic
                yield ""
                continue
            yield msg_store.render_message(x["msgid"], lang_code, x.get("params", {}))


class PackageManifest:
    def __init__(
        self,
        doc: tomlkit.TOMLDocument | InputPackageManifestType,
    ) -> None:
        self._raw_doc = doc if isinstance(doc, tomlkit.TOMLDocument) else None
        self._data = _translate_to_manifest_v1(cast(InputPackageManifestType, doc))
        if "kind" not in self._data:
            self._data["kind"] = [k for k in ALL_PACKAGE_KINDS if k in self._data]

    @classmethod
    def load_toml(cls, stream: BinaryIO) -> "Self":
        return cls(tomlkit.load(stream))

    @classmethod
    def load_from_path(cls, p: pathlib.Path) -> "Self":
        suffix = p.suffix.lower()
        match suffix:
            case ".toml":
                with open(p, "rb") as fp:
                    return cls.load_toml(fp)
            case _:
                raise RuntimeError(
                    f"unrecognized package manifest file extension: '{p.suffix}'"
                )

    def to_raw(self) -> PackageManifestType:
        return deepcopy(self._data)

    @property
    def raw_doc(self) -> tomlkit.TOMLDocument | None:
        return self._raw_doc

    @property
    def slug(self) -> str | None:
        return self._data["metadata"].get("slug")

    @property
    def kind(self) -> list[PackageKind]:
        return self._data["kind"]

    def has_kind(self, k: PackageKind) -> bool:
        return k in self._data["kind"]

    @property
    def desc(self) -> str:
        return self._data["metadata"]["desc"]

    @property
    def doc_uri(self) -> str | None:
        return self._data["metadata"].get("doc_uri")

    @property
    def vendor_name(self) -> str:
        return self._data["metadata"]["vendor"]["name"]

    @property
    def upstream_version(self) -> str | None:
        return self._data["metadata"].get("upstream_version")

    # TODO: vendor_eula

    @property
    def service_level(self) -> PackageServiceLevel:
        return PackageServiceLevel(self._data["metadata"].get("service_level", []))

    @cached_property
    def distfiles(self) -> dict[str, DistfileDecl]:
        return {x["name"]: DistfileDecl(x) for x in self._data["distfiles"]}

    @cached_property
    def binary_metadata(self) -> BinaryDecl | None:
        if not self.has_kind("binary"):
            return None
        if "binary" not in self._data:
            return None
        return BinaryDecl(self._data["binary"])

    @cached_property
    def blob_metadata(self) -> BlobDecl | None:
        if not self.has_kind("blob"):
            return None
        if "blob" not in self._data:
            return None
        return BlobDecl(self._data["blob"])

    @cached_property
    def source_metadata(self) -> SourceDecl | None:
        if not self.has_kind("source"):
            return None
        if "source" not in self._data:
            return None
        return SourceDecl(self._data["source"])

    @cached_property
    def toolchain_metadata(self) -> ToolchainDecl | None:
        if not self.has_kind("toolchain"):
            return None
        if "toolchain" not in self._data:
            return None
        return ToolchainDecl(self._data["toolchain"])

    @cached_property
    def emulator_metadata(self) -> EmulatorDecl | None:
        if not self.has_kind("emulator"):
            return None
        if "emulator" not in self._data:
            return None
        return EmulatorDecl(self._data["emulator"])

    @cached_property
    def provisionable_metadata(self) -> ProvisionableDecl | None:
        if not self.has_kind("provisionable"):
            return None
        if "provisionable" not in self._data:
            return None
        return ProvisionableDecl(self._data["provisionable"])


class BoundPackageManifest(PackageManifest):
    def __init__(
        self,
        category: str,
        name: str,
        ver: str,
        data: InputPackageManifestType,
        repo: "MetadataRepo",
    ) -> None:
        super().__init__(data)

        self.category = category
        self.name = name
        self.ver = ver
        self._semver = Version.parse(ver)
        self.repo = repo

    @property
    def repo_id(self) -> str:
        return self.repo.repo_id

    @property
    def semver(self) -> Version:
        return self._semver

    @property
    def is_prerelease(self) -> bool:
        return is_prerelease(self._semver)

    @property
    def name_for_installation(self) -> str:
        return f"{self.name}-{self.ver}"


PRERELEASE_TAGS_RE: Final = re.compile(r"^(?:alpha|beta|pre|rc)")


def is_prerelease(sv: Version) -> bool:
    if sv.prerelease is None:
        return False

    # only consider "(alpha|beta|pre|rc).*" versions as prerelease, to accommodate
    # various semver "hacks" as incorporated by upstream(s), and ourselves
    # ("ruyi.YYYYMMDD" are used as ordinary datestamps that affects sorting
    # order, in contrast to build tags).
    # see https://github.com/ruyisdk/ruyi/issues/156
    return PRERELEASE_TAGS_RE.match(sv.prerelease) is not None



from contextlib import contextmanager, redirect_stderr, redirect_stdout
from importlib import resources
from dataclasses import dataclass
import io
import os
import pathlib
import sys
from typing import Generator, cast

from pygit2 import Repository
import pytest

from ruyi.cli.completion import ArgumentParser
from ruyi.cli.main import main as ruyi_main
from ruyi.config import GlobalConfig
from ruyi.log import RuyiConsoleLogger, RuyiLogger
from ruyi.ruyipkg.repo import MetadataRepo
from ruyi.utils.global_mode import EnvGlobalModeProvider, GlobalModeProvider


class RuyiFileFixtureFactory:
    def __init__(self, module: resources.Package | None = None) -> None:
        self._fixtures_dir: pathlib.Path | None = None

        if sys.version_info < (3, 12):
            assert module is not None
        # Figure out the fixtures path in a compatible way
        if module is None:
            # Python 3.12+ fallback - get the directory of this file
            self._fixtures_dir = pathlib.Path(__file__).parent
        elif isinstance(module, str):
            # Import the module and get its path
            import importlib

            mod = importlib.import_module(module)
            if mod.__file__ is not None:
                self._fixtures_dir = pathlib.Path(mod.__file__).parent
            else:
                self._fixtures_dir = pathlib.Path(__file__).parent
        else:
            self.module = module

    @contextmanager
    def path(self, *frags: str) -> Generator[pathlib.Path, None, None]:
        if self._fixtures_dir is not None:
            # directly derive the file path for better compatibility
            result_path = self._fixtures_dir
            for frag in frags:
                result_path = result_path / frag
            yield result_path
            return

        # fallback to importlib.resources
        try:
            path = resources.files(self.module)
            for frag in frags:
                path = path.joinpath(frag)
            with resources.as_file(path) as p:
                yield p
                return
        except (TypeError, FileNotFoundError):
            pass

        # final fallback - use the directory of this file
        result_path = pathlib.Path(__file__).parent
        for frag in frags:
            result_path = result_path / frag
        yield result_path

    @contextmanager
    def plugin_suite(self, suite_name: str) -> Generator[pathlib.Path, None, None]:
        if self._fixtures_dir is not None:
            # directly derive the file path for better compatibility
            result_path = self._fixtures_dir / "plugins_suites" / suite_name
            yield result_path
            return

        # fallback to importlib.resources
        try:
            path = resources.files(self.module)
            path = path.joinpath("plugins_suites").joinpath(suite_name)
            with resources.as_file(path) as p:
                yield p
                return
        except (TypeError, FileNotFoundError):
            pass

        # final fallback - use the directory of this file
        result_path = pathlib.Path(__file__).parent / "plugins_suites" / suite_name
        yield result_path


class MockGlobalModeProvider(GlobalModeProvider):
    def __init__(
        self,
        is_debug: bool = False,
        is_experimental: bool = False,
        is_porcelain: bool = False,
        is_telemetry_optout: bool = False,
        is_cli_autocomplete: bool = False,
        venv_root: str | None = None,
    ) -> None:
        self._is_debug = is_debug
        self._is_experimental = is_experimental
        self._is_porcelain = is_porcelain
        self._is_telemetry_optout = is_telemetry_optout
        self._is_cli_autocomplete = is_cli_autocomplete
        self._venv_root = venv_root

    @property
    def argv0(self) -> str:
        return "ruyi"

    @property
    def main_file(self) -> str:
        return "ruyi/__main__.py"

    @property
    def self_exe(self) -> str:
        return "ruyi"

    @property
    def is_debug(self) -> bool:
        return self._is_debug

    @property
    def is_experimental(self) -> bool:
        return self._is_experimental

    @property
    def is_packaged(self) -> bool:
        return False

    @property
    def is_porcelain(self) -> bool:
        return self._is_porcelain

    @is_porcelain.setter
    def is_porcelain(self, v: bool) -> None:
        self._is_porcelain = v

    @property
    def is_telemetry_optout(self) -> bool:
        return self._is_telemetry_optout

    @property
    def is_cli_autocomplete(self) -> bool:
        return self._is_cli_autocomplete

    @property
    def venv_root(self) -> str | None:
        return self._venv_root


@pytest.fixture
def ruyi_file() -> RuyiFileFixtureFactory:
    return RuyiFileFixtureFactory(None if sys.version_info >= (3, 12) else __name__)


@pytest.fixture
def mock_gm() -> MockGlobalModeProvider:
    return MockGlobalModeProvider()


@pytest.fixture
def ruyi_logger(mock_gm: GlobalModeProvider) -> RuyiLogger:
    """Fixture for creating a RuyiLogger instance."""
    return RuyiConsoleLogger(mock_gm)


@dataclass
class CLIRunResult:
    exit_code: int
    stdout: str
    stderr: str


@dataclass
class CLICommandContext:
    argv: list[str]
    gm: EnvGlobalModeProvider
    gc: GlobalConfig
    stdout: io.StringIO
    stderr: io.StringIO

    @property
    def fatal_messages(self) -> list[str]:
        prefix = "fatal error: "
        return [
            line[len(prefix) :]
            for line in self.stderr.getvalue().splitlines()
            if line.startswith(prefix)
        ]


class MockRepository:
    def __init__(self, root: pathlib.Path) -> None:
        self.workdir = root
        self.path = root


class IntegrationTestHarness:
    def __init__(
        self,
        env: dict[str, str],
        repo_root: pathlib.Path,
        repo_url: str,
        repo_branch: str,
    ) -> None:
        self._env = env
        self.repo_root = repo_root
        self.repo_url = repo_url
        self.repo_branch = repo_branch

    def __call__(self, *args: str) -> CLIRunResult:
        return self.run(*args)

    def make_parser(self) -> ArgumentParser:
        return ArgumentParser()

    def make_command_context(self, *args: str) -> CLICommandContext:
        argv = ["ruyi", *args]
        stdout_io = io.StringIO()
        stderr_io = io.StringIO()
        gm = EnvGlobalModeProvider(self._env, argv)
        gm.record_self_exe(argv[0], __file__, argv[0])
        logger = RuyiConsoleLogger(gm, stdout=stdout_io, stderr=stderr_io)
        gc = GlobalConfig.load_from_config(gm, logger)
        gc.override_repo_dir = str(self.repo_root)
        gc.override_repo_url = self.repo_url
        gc.override_repo_branch = self.repo_branch
        return CLICommandContext(argv, gm, gc, stdout_io, stderr_io)

    def run(self, *args: str) -> CLIRunResult:
        ctx = self.make_command_context(*args)
        with redirect_stdout(ctx.stdout), redirect_stderr(ctx.stderr):
            exit_code = ruyi_main(ctx.gm, ctx.gc, ctx.argv)
        return CLIRunResult(
            exit_code,
            ctx.stdout.getvalue(),
            ctx.stderr.getvalue(),
        )

    def add_package(
        self,
        category: str,
        name: str,
        version: str,
        manifest_toml: str,
    ) -> pathlib.Path:
        pkg_dir = self.repo_root / "packages" / category / name
        pkg_dir.mkdir(parents=True, exist_ok=True)
        manifest_path = pkg_dir / f"{version}.toml"
        manifest_path.write_text(manifest_toml, encoding="utf-8")
        return manifest_path


def _populate_default_packages_index(repo_root: pathlib.Path) -> None:
    repo_root.mkdir(parents=True, exist_ok=True)

    config_text = """\
ruyi-repo = "v1"

[[mirrors]]
id = "ruyi-dist"
urls = ["https://example.invalid/dist/"]
"""

    (repo_root / "config.toml").write_text(config_text + "\n", encoding="utf-8")

    sha_stub = "0" * 64
    manifest_text = f"""\
format = "v1"

[metadata]
desc = "Sample integration package"
vendor = {{ name = "Ruyi Integration Tests", eula = "" }}

[[distfiles]]
name = "sample-src.tar.zst"
size = 0

[distfiles.checksums]
sha256 = "{sha_stub}"

[source]
distfiles = ["sample-src.tar.zst"]
"""

    manifest_dir = repo_root / "packages" / "dev-tools" / "sample-cli"
    manifest_dir.mkdir(parents=True, exist_ok=True)
    (manifest_dir / "1.0.0.toml").write_text(manifest_text, encoding="utf-8")


@pytest.fixture
def ruyi_cli_runner(
    tmp_path: pathlib.Path,
    monkeypatch: pytest.MonkeyPatch,
) -> IntegrationTestHarness:
    base_dir = tmp_path / "integration-env"
    home_dir = base_dir / "home"
    cache_dir = base_dir / "cache"
    config_dir = base_dir / "config"
    data_dir = base_dir / "data"
    state_dir = base_dir / "state"

    for p in (home_dir, cache_dir, config_dir, data_dir, state_dir):
        p.mkdir(parents=True, exist_ok=True)

    monkeypatch.setenv("HOME", str(home_dir))
    monkeypatch.setenv("XDG_CACHE_HOME", str(cache_dir))
    monkeypatch.setenv("XDG_CONFIG_HOME", str(config_dir))
    monkeypatch.setenv("XDG_DATA_HOME", str(data_dir))
    monkeypatch.setenv("XDG_STATE_HOME", str(state_dir))
    monkeypatch.setenv("RUYI_TELEMETRY_OPTOUT", "1")

    repo_root = cache_dir / "ruyi" / "packages-index"
    _populate_default_packages_index(repo_root)

    def _ensure_git_repo_stub(self: MetadataRepo) -> Repository:
        if self.repo is None:
            repo_path = pathlib.Path(self.root)
            repo_path.mkdir(parents=True, exist_ok=True)
            self.repo = cast(Repository, MockRepository(repo_path))
        return self.repo

    monkeypatch.setattr(MetadataRepo, "ensure_git_repo", _ensure_git_repo_stub)

    env = dict(os.environ)

    return IntegrationTestHarness(
        env=env,
        repo_root=repo_root,
        repo_url="https://example.invalid/packages-index.git",
        repo_branch="main",
    )



from io import BytesIO
import gettext
import os
from typing import Final, LiteralString, Mapping, NewType

from ..resource_bundle import get_resource_blob


def _probe_lang(environ: Mapping[str, str]) -> list[str]:
    """Probe the environment variables the gettext way, to determine the list
    of preferred languages."""
    languages: list[str] = []
    # check the variables in this order
    for envar in ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG"):
        if val := environ.get(envar):
            languages = val.split(":")
            break
    if "C" not in languages:
        languages.append("C")

    for i, lang in enumerate(languages):
        # normalize things like en_US.UTF-8 to en_US
        if "." in lang:
            languages[i] = lang.split(".", 1)[0]

    return languages


_DOMAINS = (
    "argparse",
    "ruyi",
)
"""gettext domains we supply and use ourselves"""


class I18nAdapter:
    """Adapter for gettext translation functions."""

    def __init__(self) -> None:
        self._t = gettext.NullTranslations()

    def hook(self) -> None:
        # monkey-patch the global gettext functions
        # the type ignore comments are necessary because mypy doesn't see
        # the bounded methods as compatible with the unbound functions
        # (it doesn't remove self from the unbound method signature)
        gettext.gettext = self.gettext  # type: ignore[assignment]
        gettext.ngettext = self.ngettext  # type: ignore[assignment]

    def init_from_env(self, environ: Mapping[str, str] | None = None) -> None:
        if environ is None:
            environ = os.environ

        langs = _probe_lang(environ)
        for domain in _DOMAINS:
            for lang in langs:
                if self.set_locale(domain, lang):
                    break

    def _get_mo(self, domain: str, locale: str) -> BytesIO | None:
        # this is always forward-slash-separated, because this is not a concrete
        # filesystem path, rather a resource bundle key
        path = f"locale/{locale}/LC_MESSAGES/{domain}.mo"
        blob = get_resource_blob(path)
        if blob:
            return BytesIO(blob)
        return None

    def set_locale(self, domain: str, locale: str | None = None) -> bool:
        if locale is not None:
            if mo_file := self._get_mo(domain, locale):
                self._t.add_fallback(gettext.GNUTranslations(mo_file))
                return True
        return False

    def gettext(self, x: str) -> str:
        return self._t.gettext(x)

    def ngettext(self, singular: str, plural: str, n: int) -> str:
        return self._t.ngettext(singular, plural, n)


ADAPTER: Final = I18nAdapter()


DeferredI18nString = NewType("DeferredI18nString", str)


def _(x: LiteralString | DeferredI18nString) -> str:
    """``gettext`` alias that ensures its input is string literal via type
    signature."""
    return ADAPTER.gettext(x)


def d_(x: LiteralString) -> DeferredI18nString:
    """Mark a string literal for deferred translation: call ``_`` at use sites."""
    return DeferredI18nString(x)



from typing import Any, Mapping, Sequence

from ruyi.log import RuyiLogger
from ruyi.ruyipkg.entity_provider import BaseEntityProvider, FSEntityProvider
from ruyi.ruyipkg.entity import EntityStore

from ..fixtures import RuyiFileFixtureFactory


class MockEntityProvider(BaseEntityProvider):
    """A mock entity provider for testing."""

    def discover_schemas(self) -> dict[str, object]:
        """Return a mock schema."""
        return {
            "os": {
                "$schema": "http://json-schema.org/draft-07/schema#",
                "type": "object",
                "properties": {
                    "os": {
                        "type": "object",
                        "properties": {
                            "display_name": {"type": "string"},
                            "version": {"type": "string"},
                        },
                        "required": ["display_name"],
                    }
                },
            }
        }

    def load_entities(
        self,
        entity_types: Sequence[str],
    ) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
        """Return mock entity data if 'os' is in entity_types."""
        if "os" not in entity_types:
            return {}

        return {
            "os": {
                "linux": {"os": {"display_name": "Linux", "version": "6.6.0"}},
                "freebsd": {"os": {"display_name": "FreeBSD", "version": "14.0"}},
            }
        }


def test_entity_store_with_custom_provider(
    ruyi_file: RuyiFileFixtureFactory,
    ruyi_logger: RuyiLogger,
) -> None:
    """Test using EntityStore with a custom provider."""

    with ruyi_file.path("ruyipkg_suites", "entities_v0_smoke") as entities_path:
        # Create store with both filesystem and mock providers
        fs_provider = FSEntityProvider(ruyi_logger, entities_path)
        mock_provider = MockEntityProvider()
        store = EntityStore(ruyi_logger, fs_provider, mock_provider)

        # Verify entity types from both providers are available
        entity_types = set(store.get_entity_types())
        assert "cpu" in entity_types  # from filesystem
        assert "os" in entity_types  # from mock provider

        # Verify we can get entities from both providers
        cpu = store.get_entity("cpu", "xiangshan-nanhu")
        assert cpu is not None
        assert cpu.entity_type == "cpu"

        os = store.get_entity("os", "linux")
        assert os is not None
        assert os.entity_type == "os"
        assert os.display_name == "Linux"
        assert os.data.get("version") == "6.6.0"



"""Executor for ``ruyi admin build-package``.

Given a path to a recipe ``.star`` file, this module:

1. Discovers the enclosing recipe project (via :mod:`recipe_project`).
2. Builds a :class:`PluginHostContext` in build-recipe mode.
3. Loads the recipe module, collecting :class:`ScheduledBuild`
   registrations.
4. For each (optionally ``--name``-filtered) scheduled build, constructs
   a :class:`RecipeBuildCtx` and calls the registered callable to obtain
   one or more :class:`Invocation` plans.
5. Executes each plan via :func:`subprocess.run` unless ``dry_run`` is
   set; on a zero exit, resolves declared ``produces`` globs, computes
   sha256/sha512/size for each matched file, and writes a build-report
   TOML.

The runner is intentionally I/O-heavy at call time and light on imports
at the module top level so that the CLI startup-flow lint stays happy;
all subprocess/hashlib/tomllib usage is inside :func:`run_recipe`.
"""

from __future__ import annotations

from dataclasses import dataclass
import pathlib
from typing import TYPE_CHECKING, Any, Iterable, Mapping, Sequence

from ..log import RuyiLogger
from ..pluginhost.build_api import (
    Invocation,
    RecipeBuildCtx,
    ScheduledBuild,
)
from .recipe_project import RecipeProject, discover_recipe_project

if TYPE_CHECKING:
    from ..pluginhost.ctx import PluginHostContext


@dataclass(frozen=True)
class ArtifactReport:
    path: pathlib.Path
    size: int
    checksums: Mapping[str, str]


@dataclass(frozen=True)
class BuildReport:
    recipe_file: pathlib.Path
    project_name: str
    build_name: str
    invocations: tuple[Invocation, ...]
    artifacts: tuple[ArtifactReport, ...]
    exit_code: int


class BuildFailure(RuntimeError):
    """Raised when an Invocation exits non-zero."""

    def __init__(self, build_name: str, exit_code: int) -> None:
        super().__init__(f"build {build_name!r} failed with exit code {exit_code}")
        self.build_name = build_name
        self.exit_code = exit_code


def run_recipe(
    logger: RuyiLogger,
    recipe_file: pathlib.Path,
    *,
    user_vars: Mapping[str, str] | None = None,
    selected_names: Sequence[str] | None = None,
    dry_run: bool = False,
    output_dir_override: pathlib.Path | None = None,
) -> list[BuildReport]:
    """Load the recipe and execute (or plan, if ``dry_run``) each selected
    scheduled build.

    Returns a :class:`BuildReport` per executed build. Raises
    :class:`BuildFailure` on the first non-zero exit (subsequent builds
    are not attempted).
    """

    project = discover_recipe_project(recipe_file)
    if output_dir_override is not None:
        project = _with_output_dir(project, output_dir_override)

    phctx = _make_recipe_phctx(logger, project)
    scheduled = _load_recipe_module(phctx, recipe_file)
    if not scheduled:
        raise RuntimeError(
            f"recipe {recipe_file} scheduled no builds "
            f"(did you call RUYI.build.schedule_build(...)?)"
        )

    if selected_names is not None:
        selected = _filter_by_name(scheduled, selected_names)
    else:
        selected = list(scheduled)

    reports: list[BuildReport] = []
    for sb in selected:
        reports.append(
            _execute_one_build(
                logger,
                phctx,
                project,
                sb,
                user_vars=user_vars or {},
                dry_run=dry_run,
            )
        )
    return reports


def _with_output_dir(
    project: RecipeProject, new_output_dir: pathlib.Path
) -> RecipeProject:
    resolved = new_output_dir.resolve()
    return RecipeProject(
        root=project.root,
        name=project.name,
        output_dir=resolved,
        extra_artifact_roots=project.extra_artifact_roots,
    )


def _make_recipe_phctx(
    logger: RuyiLogger, project: RecipeProject
) -> "PluginHostContext[Any, Any]":
    # Local import: keeps heavy module-load chains out of CLI startup.
    from ..pluginhost.ctx import PluginHostContext

    return PluginHostContext.new(
        logger,
        project.root,  # plugin_root is reused for the recipe project
        recipe_project_root=project.root,
    )


def _load_recipe_module(
    phctx: "PluginHostContext[Any, Any]", recipe_file: pathlib.Path
) -> list[ScheduledBuild]:
    resolved = recipe_file.resolve(strict=True)
    return phctx.load_recipe(resolved)


def _filter_by_name(
    scheduled: Iterable[ScheduledBuild], wanted: Sequence[str]
) -> list[ScheduledBuild]:
    by_name = {sb.name: sb for sb in scheduled}
    missing = [n for n in wanted if n not in by_name]
    if missing:
        raise RuntimeError(
            f"recipe does not define the requested build(s): {', '.join(missing)}"
        )
    return [by_name[n] for n in wanted]


def _execute_one_build(
    logger: RuyiLogger,
    phctx: "PluginHostContext[Any, Any]",
    project: RecipeProject,
    sb: ScheduledBuild,
    *,
    user_vars: Mapping[str, str],
    dry_run: bool,
) -> BuildReport:
    ctx = RecipeBuildCtx(
        project=project,
        name=sb.name,
        recipe_file=sb.recipe_file,
        user_vars=user_vars,
    )
    ev = phctx.make_evaluator()
    result = ev.eval_function(sb.fn, ctx)
    invocations = _normalize_invocations(sb.name, result)

    if dry_run:
        for inv in invocations:
            logger.I(f"[dry-run] would run: {' '.join(inv.argv)} (cwd={inv.cwd})")
        return BuildReport(
            recipe_file=sb.recipe_file,
            project_name=project.name,
            build_name=sb.name,
            invocations=tuple(invocations),
            artifacts=(),
            exit_code=0,
        )

    last_exit = 0
    for inv in invocations:
        last_exit = _run_invocation(logger, inv)
        if last_exit != 0:
            raise BuildFailure(sb.name, last_exit)

    artifact_reports = _resolve_artifacts(invocations)

    return BuildReport(
        recipe_file=sb.recipe_file,
        project_name=project.name,
        build_name=sb.name,
        invocations=tuple(invocations),
        artifacts=tuple(artifact_reports),
        exit_code=last_exit,
    )


def _normalize_invocations(build_name: str, result: object) -> list[Invocation]:
    if isinstance(result, Invocation):
        return [result]
    if isinstance(result, (list, tuple)):
        out: list[Invocation] = []
        for item in result:
            if not isinstance(item, Invocation):
                raise RuntimeError(
                    f"build {build_name!r}: expected Invocation (from "
                    f"ctx.subprocess), got {type(item).__name__}"
                )
            out.append(item)
        if not out:
            raise RuntimeError(
                f"build {build_name!r}: returned an empty list of Invocations"
            )
        return out
    raise RuntimeError(
        f"build {build_name!r}: expected Invocation or list of Invocations, "
        f"got {type(result).__name__}"
    )


def _run_invocation(logger: RuyiLogger, inv: Invocation) -> int:
    import os
    import subprocess

    logger.I(f"running: {' '.join(inv.argv)} (cwd={inv.cwd})")
    merged_env = os.environ.copy()
    merged_env.update(inv.env)

    # SECURITY: recipes are explicit Starlark code evaluated in a trusted recipe
    # project, and the invocation takes an argv list with no shell expansion,
    # which is the documented threat model.
    proc = subprocess.run(
        list(inv.argv),  # sourcery skip: dangerous-subprocess-use-audit
        cwd=str(inv.cwd),
        env=merged_env,
        check=False,
    )
    return proc.returncode


def _resolve_artifacts(
    invocations: Iterable[Invocation],
) -> list[ArtifactReport]:
    import os

    from . import checksum

    reports: list[ArtifactReport] = []
    for inv in invocations:
        for art in inv.produces:
            matches = sorted(art.root.glob(art.glob))
            if not matches:
                raise RuntimeError(
                    f"artifact {art.glob!r} under {art.root} matched no files"
                )
            for match in matches:
                if not match.is_file():
                    continue
                with open(match, "rb") as fp:
                    size = os.stat(fp.fileno()).st_size
                    csums = checksum.Checksummer(fp, {}).compute(
                        kinds=checksum.SUPPORTED_CHECKSUM_KINDS,
                    )
                reports.append(
                    ArtifactReport(
                        path=match,
                        size=size,
                        checksums=csums,
                    )
                )
    return reports


def format_build_report(report: BuildReport) -> str:
    """Return a TOML-shaped serialization of a build report."""

    lines: list[str] = []
    lines.append("# ruyi build report")
    lines.append(f'recipe_file = "{report.recipe_file}"')
    lines.append(f'project_name = "{report.project_name}"')
    lines.append(f'build_name = "{report.build_name}"')
    lines.append(f"exit_code = {report.exit_code}")
    for inv in report.invocations:
        lines.append("")
        lines.append("[[invocations]]")
        lines.append(f"argv = {list(inv.argv)!r}")
        lines.append(f'cwd = "{inv.cwd}"')
        if inv.env:
            lines.append(f"env = {dict(inv.env)!r}")
    for art in report.artifacts:
        lines.append("")
        lines.append("[[artifacts]]")
        lines.append(f'path = "{art.path}"')
        lines.append(f"size = {art.size}")
        for kind in sorted(art.checksums):
            lines.append(f'{kind} = "{art.checksums[kind]}"')
    return "\n".join(lines) + "\n"



from ruyi.utils import mounts


def test_parse_mounts() -> None:
    sample = r"""
/dev/mapper/foo / btrfs rw,noatime,ssd,discard=async,space_cache=v2,autodefrag,subvolid=5,subvol=/ 0 0
devtmpfs /dev devtmpfs rw,nosuid,size=3896808k,nr_inodes=974202,mode=755,inode64 0 0
tmpfs /dev/shm tmpfs rw,nosuid,nodev,inode64 0 0
tmpfs /tmp/x\040b tmpfs rw,relatime,inode64 0 0
"""

    parsed = mounts.parse_mounts(sample)
    assert len(parsed) == 4

    assert parsed[0].source == "/dev/mapper/foo"
    assert parsed[0].target == "/"
    assert parsed[0].fstype == "btrfs"
    assert parsed[0].options == [
        "rw",
        "noatime",
        "ssd",
        "discard=async",
        "space_cache=v2",
        "autodefrag",
        "subvolid=5",
        "subvol=/",
    ]

    assert parsed[3].source == "tmpfs"
    assert parsed[3].target == "/tmp/x b"
    assert parsed[3].fstype == "tmpfs"
    assert parsed[3].options == ["rw", "relatime", "inode64"]



#!/usr/bin/env python3

import ast
import os
import sys
import tomllib


def main() -> None:
    # assume CWD is project root, which is guaranteed to be the case (see
    # end of file)
    with open("pyproject.toml", "rb") as fp:
        poetry2_project = tomllib.load(fp)
        poetry2_version = poetry2_project["project"]["version"]

    with open("contrib/poetry-1.x/pyproject.toml", "rb") as fp:
        poetry1_project = tomllib.load(fp)
        poetry1_version = poetry1_project["tool"]["poetry"]["version"]

    if poetry1_version != poetry2_version:
        print("fatal error: Poetry 1.x metadata inconsistent with primary data source")
        print(f"info: primary pyproject.toml has project.version = '{poetry2_version}'")
        print(f"info: Poetry 1.x has tool.poetry.version = '{poetry1_version}'")
        sys.exit(1)

    print("info: project version consistent between primary and Poetry 1.x metadata")

    ret = lint_ruyi_version_str("ruyi/version.py", poetry2_version)
    if ret:
        print(
            "info: hint, if you want to refactor RUYI_SEMVER, you need to make changes to this lint too"
        )
        sys.exit(ret)


def lint_ruyi_version_str(filename: str, expected_ver: str) -> int:
    with open(filename, "rb") as fp:
        contents = fp.read()

    module = ast.parse(contents, filename)
    found_ver: str | None = None
    for stmt in module.body:
        if not isinstance(stmt, ast.AnnAssign):
            continue
        if not isinstance(stmt.target, ast.Name):
            continue
        if stmt.target.id == "RUYI_SEMVER":
            if not isinstance(stmt.value, ast.Constant):
                print("fatal error: RUYI_SEMVER not a constant")
                return 1
            if not isinstance(stmt.value.value, str):
                print("fatal error: RUYI_SEMVER not a string")
                return 1
            found_ver = stmt.value.value

    if found_ver is None:
        print("fatal error: RUYI_SEMVER annotation assignment not found")
        return 1

    if found_ver != expected_ver:
        print(
            "fatal error: ruyi.version.RUYI_SEMVER inconsistent with primary data source"
        )
        print(f"info: primary pyproject.toml has project.version = '{expected_ver}'")
        print(f"info: ruyi.version.RUYI_SEMVER = '{found_ver}'")
        return 1

    print("info: ruyi.version.RUYI_SEMVER consistent with primary metadata")
    return 0


if __name__ == "__main__":
    # cd to project root
    os.chdir(os.path.join(os.path.dirname(__file__), ".."))
    main()



import argparse
import pathlib
from typing import TYPE_CHECKING, cast

from ..cli.cmd import AdminCommand
from ..i18n import _

if TYPE_CHECKING:
    from ..cli.completion import ArgumentParser
    from ..config import GlobalConfig


class AdminCheckCommand(
    AdminCommand,
    cmd="check",
    help=_("Check package manifests and metadata repositories"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        input_group = p.add_mutually_exclusive_group(required=True)
        input_group.add_argument(
            "-f",
            "--file",
            action="append",
            default=None,
            metavar="MANIFEST.toml",
            help=_("Path to a package manifest to check (repeatable)"),
        )
        input_group.add_argument(
            "--repo",
            type=str,
            default=None,
            metavar="REPO_ROOT",
            help=_("Path to a metadata repository root to check"),
        )
        p.add_argument(
            "--check",
            action="append",
            choices=["format", "parse"],
            default=None,
            help=_("Check to run (repeatable; defaults to all checks)"),
        )
        p.add_argument(
            "--only-packages",
            nargs=argparse.REMAINDER,
            default=None,
            metavar="RUYI_LIST_FILTER",
            help=_(
                "Only check packages matching trailing ruyi list filters; "
                "valid only with --repo"
            ),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .check import CheckUsageError
        from .check_cli import do_admin_check

        files: list[str] | None = args.file
        repo: str | None = args.repo
        checks: list[str] | None = args.check
        only_packages: list[str] | None = args.only_packages

        if only_packages is not None and repo is None:
            cfg.logger.F(_("--only-packages is only valid with --repo"))
            return 1

        try:
            return do_admin_check(
                cfg,
                files=files,
                repo=repo,
                checks=checks,
                only_packages=only_packages,
            )
        except CheckUsageError as exc:
            cfg.logger.F(str(exc))
            return 1


class AdminChecksumCommand(
    AdminCommand,
    cmd="checksum",
    help=_("Generate a checksum section for a manifest file for the distfiles given"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "--format",
            "-f",
            type=str,
            choices=["toml"],
            default="toml",
            help=_("Format of checksum section to generate in"),
        )
        p.add_argument(
            "--restrict",
            type=str,
            default="",
            help=_(
                "the 'restrict' field to use for all mentioned distfiles, separated with comma"
            ),
        )
        p.add_argument(
            "file",
            type=str,
            nargs="+",
            help=_("Path to the distfile(s) to checksum"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .admin_checksum import do_admin_checksum

        logger = cfg.logger
        files = args.file
        format = args.format
        restrict_str = cast(str, args.restrict)
        restrict = restrict_str.split(",") if restrict_str else []

        return do_admin_checksum(logger, files, format, restrict)


class AdminFormatManifestCommand(
    AdminCommand,
    cmd="format-manifest",
    help=_("Format the given package manifests into canonical TOML representation"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "file",
            type=str,
            nargs="+",
            help=_("Path to the distfile(s) to generate manifest for"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .manifest_io import dump_canonical_package_manifest_from_path

        files = args.file

        for f in files:
            p = pathlib.Path(f)
            d = dump_canonical_package_manifest_from_path(p)

            dest_path = p.with_suffix(".toml")
            with open(dest_path, "w", encoding="utf-8") as fp:
                fp.write(d)

        return 0


class AdminBuildPackageCommand(
    AdminCommand,
    cmd="build-package",
    help=_("Build a package from a recipe file"),
):
    @classmethod
    def configure_args(cls, gc: "GlobalConfig", p: "ArgumentParser") -> None:
        p.add_argument(
            "recipe_file",
            type=str,
            help=_("Path to the recipe .star file"),
        )
        p.add_argument(
            "-v",
            "--var",
            action="append",
            default=[],
            metavar="KEY=VALUE",
            help=_("Set a user variable for the recipe (repeatable)"),
        )
        p.add_argument(
            "-n",
            "--name",
            action="append",
            default=[],
            metavar="NAME",
            help=_(
                "Select a specific scheduled build by name (repeatable); "
                "by default all scheduled builds are executed"
            ),
        )
        p.add_argument(
            "--dry-run",
            action="store_true",
            help=_("Print the build plan without executing it"),
        )
        p.add_argument(
            "--output-dir",
            type=str,
            default=None,
            help=_("Override the recipe project's output directory"),
        )

    @classmethod
    def main(cls, cfg: "GlobalConfig", args: argparse.Namespace) -> int:
        from .build_runner import (
            BuildFailure,
            format_build_report,
            run_recipe,
        )

        logger = cfg.logger
        recipe_file = pathlib.Path(cast(str, args.recipe_file))
        var_strs = cast("list[str]", args.var)
        selected_names = cast("list[str]", args.name) or None
        dry_run = cast(bool, args.dry_run)
        output_dir_raw = cast("str | None", args.output_dir)
        output_dir = (
            pathlib.Path(output_dir_raw) if output_dir_raw is not None else None
        )

        user_vars: dict[str, str] = {}
        for v in var_strs:
            if "=" not in v:
                logger.F(
                    _("invalid --var spec {spec!r}: expected KEY=VALUE").format(
                        spec=v,
                    )
                )
                return 1
            k, _sep, val = v.partition("=")
            if not k:
                logger.F(_("invalid --var spec {spec!r}: empty key").format(spec=v))
                return 1
            user_vars[k] = val

        try:
            reports = run_recipe(
                logger,
                recipe_file,
                user_vars=user_vars,
                selected_names=selected_names,
                dry_run=dry_run,
                output_dir_override=output_dir,
            )
        except BuildFailure as e:
            logger.F(str(e))
            return e.exit_code or 1
        except (RuntimeError, FileNotFoundError) as e:
            logger.F(str(e))
            return 1

        for r in reports:
            logger.I(
                _("build {name!r} completed: {n} artifact(s)").format(
                    name=r.build_name,
                    n=len(r.artifacts),
                )
            )
            print(format_build_report(r))

        return 0
